# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A base markdown node."""
from __future__ import annotations

import logging
import math
import re
from abc import abstractmethod
from typing import Dict, Iterable, List, Optional, Tuple

import trestle.core.markdown.markdown_const as md_const
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_filtered_list, delete_list_from_list

logger = logging.getLogger(__name__)


class BaseSectionContent:
    """A content of the node."""

    def __init__(self) -> None:
        """Initialize section content."""
        self.raw_text = ''
        self.subnodes_keys: List[str] = []

    def union(self, node: BaseMarkdownNode) -> None:
        """Unites contents together."""
        self.subnodes_keys.append(node.key)
        self.subnodes_keys.extend(node.content.subnodes_keys)


class BaseMarkdownNode:
    """Markdown will be read to the tree."""

    def __init__(self, key: str, content: BaseSectionContent, starting_line: int):
        """Initialize markdown node."""
        self.subnodes: List[BaseMarkdownNode] = []
        self.key = key
        self.content = content
        self.starting_line = starting_line

    @classmethod
    def build_tree_from_markdown(cls, lines: List[str]) -> BaseMarkdownNode:
        """Construct a tree out of the given markdown."""
        ob = cls.__new__(cls)
        start_level = ob._get_max_header_lvl(lines)
        ob, _ = ob._build_tree(lines, 'root', 0, start_level)
        return ob

    def get_all_headers_for_level(self, level: int) -> Iterable[str]:
        """Return all headers per specified level of hierarchy."""
        return list(
            filter(lambda header: self._get_header_level_if_valid(header) == level, self.content.subnodes_keys)
        ).__iter__()

    def get_node_for_key(self, key: str, strict_matching: bool = True) -> Optional[BaseMarkdownNode]:
        """Return a first node for the given key, substring matching is supported. The method is case insensitive."""
        if not strict_matching:
            if not any(key.lower() in el.lower() for el in self.content.subnodes_keys):
                return None
            elif len(as_filtered_list(self.content.subnodes_keys, lambda el: key.lower() in el.lower())) > 1:
                logger.warning(f'Multiple nodes for {key} were found, only the first one will be returned.')
        else:
            if key.lower() not in [el.lower() for el in self.content.subnodes_keys]:
                return None
            elif len(as_filtered_list(self.content.subnodes_keys, lambda el: el.lower() == key.lower())) > 1:
                logger.warning(f'Multiple nodes for {key} were found, only the first one will be returned.')

        return self._rec_traverse(self, key, strict_matching)

    def get_all_nodes_for_keys(
        self,
        keys: List[str],
        strict_matching: bool = True,
        stop_recurse_on_first_match: bool = False
    ) -> List[BaseMarkdownNode]:
        """
        Return all nodes for the given keys, substring matching is supported.

        Args:
            keys: List of strings for the headers being collected
            strict_matching: Force exact match of key with header vs. simple substring match
            stop_recurse_on_first_match: Return first match of any of the keys and don't search subnodes

        Returns: List of found markdown nodes
        """
        if not strict_matching:
            if not any(key in el for el in self.content.subnodes_keys for key in keys):
                return []
        elif not set(keys).intersection(self.content.subnodes_keys):
            return []

        return self._rec_traverse_all(self, keys, strict_matching, stop_recurse_on_first_match)

    def get_all_headers_for_key(self, key: str, strict_matching: bool = True) -> Iterable[str]:
        """Return all headers contained in the node with a given key."""
        if strict_matching:
            return list(filter(lambda header: key == header, self.content.subnodes_keys)).__iter__()
        else:
            return list(filter(lambda header: key in header, self.content.subnodes_keys)).__iter__()

    def get_node_header_lvl(self) -> Optional[int]:
        """Return current node header level."""
        return self._get_header_level_if_valid(self.key)

    def change_header_level_by(self, delta_level: int) -> None:
        """
        Change all headers in the tree by specified level up or down.

        All children nodes will be modified by specified level as well.

        Args:
            delta_level: each header will be modified by this number, can be negative.
        """
        # construct a map
        header_map = {}
        if self.key != 'root':
            new_key = self._modify_header_level(self.key, delta_level)
            header_map[self.key] = new_key
        for key in self.content.subnodes_keys:
            new_key = self._modify_header_level(key, delta_level)
            header_map[key] = new_key

        # go through all contents and modify headers
        self._rec_traverse_header_update(self, header_map)

    def delete_nodes_text(self, keys: List[str], strict_matching: bool = True) -> List[str]:
        """Remove text from this node that is found in matching subnodes."""
        text_lines = self.content.raw_text.split('\n')
        matching_nodes = self.get_all_nodes_for_keys(keys, strict_matching, True)
        # need to delete from end and proceed backwards
        sorted_nodes = sorted(matching_nodes, key=lambda node: node.starting_line, reverse=True)
        for node in sorted_nodes:
            last_line = node.starting_line + len(node.content.raw_text.split('\n'))
            delete_list_from_list(text_lines, list(range(node.starting_line, last_line)))
        return text_lines

    @abstractmethod
    def _build_tree(self, lines: List[str], root_key: str, starting_line: int,
                    level: int) -> Tuple[BaseMarkdownNode, int]:
        """Build a tree from the markdown recursively."""
        pass

    def _modify_header_level(self, header: str, delta_level: int) -> str:
        """Modify header level by specified level."""
        if delta_level == 0:
            logger.debug('Nothing to modify in header, level 0 is given.')
            return header

        current_level = self._get_header_level_if_valid(header)
        if current_level is None:
            current_level = 0
        if current_level + delta_level < 0:
            logger.warning(
                f'Cannot substract {delta_level} as level of {header} is {current_level}. All `#` will be removed.'
            )
            delta_level = current_level * -1

        if current_level + delta_level == 0:
            replacement = ''
        else:
            replacement = '#' * (current_level + delta_level)
        header = header.replace('#' * current_level, replacement)

        return header.strip(' ')

    def _get_header_level_if_valid(self, line: str) -> Optional[int]:
        """
        Return a level of the header if the given line is indeed a header.

        Level of the header is determined by the number of # symbols.
        """
        header_symbols = re.match(md_const.HEADER_REGEX, line)
        # Header is valid only if it line starts with header
        if header_symbols is not None and header_symbols.regs[0][0] == 0:
            return header_symbols.regs[0][1]
        return None

    def _does_start_with(self, line: str, start_chars: str) -> bool:
        """Determine whether the line starts with given characters."""
        return line.startswith(start_chars)

    def _does_contain(self, line: str, reg: str) -> bool:
        """Determine if the line matches regex."""
        if len(line) == 0 and reg != r'':
            return False
        regexp = re.compile(reg)
        return regexp.search(line) is not None

    def _read_code_lines(self, lines: List[str], line: str, i: int) -> tuple[list[str], int]:
        """Read code block."""
        code_lines = [line]
        while True:
            if i >= len(lines):
                raise TrestleError(f'Code block is not closed: {code_lines}')

            line = lines[i]
            code_lines.append(line)
            i += 1
            if self._does_contain(line, md_const.CODEBLOCK_DEF):
                break
        return code_lines, i

    def _read_html_block(self, lines: List[str], line: str, i: int, ending_regex: str) -> tuple[list[str], int]:
        """Read html block."""
        html_block = [line]
        if self._does_contain(line, r'<br[ /]*>'):
            return html_block, i
        if self._does_contain(line, ending_regex):
            return html_block, i
        while True:
            if i >= len(lines):
                raise TrestleError(f'HTML block is not closed: {html_block}')

            line = lines[i]
            html_block.append(line)
            i += 1
            if self._does_contain(line, ending_regex):
                break
        return html_block, i

    def _read_table_block(self, lines: List[str], line: str, i: int) -> tuple[list[str], int]:
        """Read table."""
        table_block = [line]
        while True:
            if i >= len(lines):
                return table_block, i

            line = lines[i]
            if not self._does_contain(line, md_const.TABLE_REGEX):
                table_block.append(line)
                break
            table_block.append(line)
            i += 1
        return table_block, i

    def _rec_traverse(self, node: BaseMarkdownNode, key: str, strict_matching: bool) -> Optional[BaseMarkdownNode]:
        """
        Recursevely traverses the tree and searches for the given key.

        If strict matching is turned off, node will be matched if key is a substring of the node's header.
        """
        if key.lower() == node.key.lower() or (not strict_matching and key.lower() in node.key.lower()):
            return node
        if (not strict_matching and any(key.lower() in el.lower()
                                        for el in node.content.subnodes_keys)) or (key.lower() in [
                                            el.lower() for el in node.content.subnodes_keys
                                        ]):
            for subnode in node.subnodes:
                matched_node = self._rec_traverse(subnode, key, strict_matching)
                if matched_node is not None:
                    return matched_node

        return None

    def _rec_traverse_all(
        self, node: BaseMarkdownNode, keys: List[str], strict_matching: bool, stop_recurse_on_first_match: bool
    ) -> List[BaseMarkdownNode]:
        """
        Recursevely traverse the tree and find all nodes matching the keys.

        If strict matching is turned off, nodes will be matched if key is a substring of the node's header.
        stop_recurse_on_first_match will return only the highest level key match and not any subnodes
        """
        found_nodes: List[BaseMarkdownNode] = []
        for key in keys:
            if key == node.key or (not strict_matching and key in node.key):
                found_nodes.append(node)
                if stop_recurse_on_first_match:
                    return found_nodes
        for subnode in node.subnodes:
            matched_nodes = self._rec_traverse_all(subnode, keys, strict_matching, stop_recurse_on_first_match)
            found_nodes.extend(matched_nodes)
        return found_nodes

    def _rec_traverse_header_update(self, node: BaseMarkdownNode, header_map: Dict[str, str]) -> None:
        """Recursively traverse tree and update the contents."""
        if node:
            if node.key != 'root':
                new_key = header_map[node.key]
                node.key = new_key

            # update text
            lines = node.content.raw_text.split('\n')
            if lines:
                for i in range(0, len(lines)):
                    line = lines[i]
                    if line in header_map.keys():
                        new_key = header_map[line]
                        lines[i] = new_key
                    elif line.strip(' ') in header_map.keys():
                        # keep spaces if any
                        new_key = header_map[line.strip(' ')]
                        lines[i] = line.replace(line.strip(' '), new_key)

                node.content.raw_text = '\n'.join(lines)

            # update subnodes
            if node.content.subnodes_keys:
                for i in range(0, len(node.content.subnodes_keys)):
                    subnode_key = node.content.subnodes_keys[i]
                    if subnode_key in header_map.keys():
                        new_key = header_map[subnode_key]
                        node.content.subnodes_keys[i] = new_key

        for subnode in node.subnodes:
            self._rec_traverse_header_update(subnode, header_map)

    def _get_max_header_lvl(self, lines: List[str]) -> int:
        """Go through all lines to determine highest header level. Less # means higher."""
        min_lvl = math.inf
        for line in lines:
            line = line.strip(' ')
            header_lvl = self._get_header_level_if_valid(line)

            if header_lvl is not None and header_lvl < min_lvl:
                min_lvl = header_lvl

        return min_lvl - 1



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fs module."""

import pathlib
from datetime import datetime, timedelta
from typing import List

import pytest

from tests import test_utils

import trestle.common.const as const
import trestle.oscal.common as common
from trestle.common import file_utils, trash
from trestle.common.err import TrestleError
from trestle.common.model_utils import ModelUtils
from trestle.core.models.file_content_type import FileContentType
from trestle.oscal import catalog

if file_utils.is_windows():  # pragma: no cover
    import win32api
    import win32con


def test_should_ignore() -> None:
    """Test should_ignore method."""
    assert ModelUtils._should_ignore('.test') is True
    assert ModelUtils._should_ignore('_test') is True
    assert ModelUtils._should_ignore('__test') is True
    assert ModelUtils._should_ignore('test') is False


def test_oscal_dir_valid(tmp_path: pathlib.Path) -> None:
    """Test if oscal dir is valid or not."""
    assert file_utils.check_oscal_directories(tmp_path)

    create_sample_catalog_project(tmp_path)

    assert file_utils.check_oscal_directories(tmp_path)

    # add some hidden files
    hidden_file = tmp_path / 'catalogs' / '.hidden.txt'
    test_utils.make_hidden_file(hidden_file)

    keep_file = tmp_path / 'catalogs' / '.keep'
    test_utils.make_hidden_file(keep_file)

    assert file_utils.check_oscal_directories(tmp_path)
    # bad hidden files are not removed but user is asked to remove them
    assert hidden_file.exists()
    assert keep_file.exists()

    # add some markdown readme
    readme_file = tmp_path / 'catalogs' / 'README.md'
    readme_file.touch()
    assert file_utils.check_oscal_directories(tmp_path)


def test_oscal_dir_notvalid(tmp_path: pathlib.Path) -> None:
    """Test OSCAL directory not valid."""
    assert file_utils.check_oscal_directories(tmp_path)
    create_sample_catalog_project(tmp_path)
    assert file_utils.check_oscal_directories(tmp_path)

    profiles_dir = tmp_path / 'profiles'
    profiles_dir.mkdir(parents=True, exist_ok=True)

    invalid_file = profiles_dir / 'shouldnt_be_here.txt'
    invalid_file.touch()

    assert not file_utils.check_oscal_directories(tmp_path)

    invalid_file.unlink()

    assert file_utils.check_oscal_directories(tmp_path)

    metadata_dir = tmp_path / 'catalogs' / 'mycatalog' / 'catalog' / 'metadata'
    deep_invalid_file = metadata_dir / 'responsible-parties' / 'should_be_here.docx'
    readme_file = tmp_path / 'catalogs' / 'readme.md'
    deep_invalid_file.touch()
    readme_file.touch()

    assert not file_utils.check_oscal_directories(tmp_path)


def test_is_valid_project_root(tmp_path: pathlib.Path) -> None:
    """Test is_valid_project_root method."""
    assert file_utils.is_valid_project_root(tmp_path) is False

    test_utils.ensure_trestle_config_dir(tmp_path)
    assert file_utils.is_valid_project_root(tmp_path) is True


def test_has_parent_path(tmp_path: pathlib.Path) -> None:
    """Test has_parent_path method."""
    assert trash.has_parent_path(pathlib.Path('tests'), test_utils.BASE_TMP_DIR) is False
    assert trash.has_parent_path(pathlib.Path('/invalid/path'), test_utils.BASE_TMP_DIR) is False


def test_get_trestle_project_root(tmp_path: pathlib.Path, rand_str: str) -> None:
    """Test get_trestle_project_root  method."""
    project_path: pathlib.Path = pathlib.Path.joinpath(tmp_path, rand_str)
    sub_path: pathlib.Path = project_path.joinpath('samples2')
    sub_path.mkdir(exist_ok=True, parents=True)
    assert sub_path.exists() and sub_path.is_dir()

    # create a file
    sub_path.joinpath('readme.md').touch()

    # create a data-dir and a file
    sub_data_dir = pathlib.Path.joinpath(sub_path, 'data')
    sub_data_dir.mkdir(exist_ok=True, parents=True)
    sub_data_dir.joinpath('readme.md').touch()

    assert file_utils.extract_trestle_project_root(sub_data_dir) is None

    test_utils.ensure_trestle_config_dir(project_path)
    assert file_utils.extract_trestle_project_root(sub_data_dir) == project_path
    assert file_utils.extract_trestle_project_root(sub_data_dir.joinpath('readme.md')) == project_path
    assert file_utils.extract_trestle_project_root(sub_path.joinpath('readme.md')) == project_path
    assert file_utils.extract_trestle_project_root(sub_path) == project_path
    assert file_utils.extract_trestle_project_root(project_path.parent) is None


def test_is_valid_project_model_path(tmp_path: pathlib.Path) -> None:
    """Test is_valid_project_model method."""
    assert file_utils._is_valid_project_model_path(tmp_path) is False

    test_utils.ensure_trestle_config_dir(tmp_path)
    assert file_utils._is_valid_project_model_path(tmp_path) is False

    create_sample_catalog_project(tmp_path)

    catalog_dir = tmp_path / 'catalogs'
    assert file_utils._is_valid_project_model_path(catalog_dir) is False

    mycatalog_dir = catalog_dir / 'mycatalog'
    assert file_utils._is_valid_project_model_path(mycatalog_dir) is True

    metadata_dir = mycatalog_dir / 'metadata'
    assert file_utils._is_valid_project_model_path(metadata_dir) is True

    foo_dir = tmp_path / 'foo/bar'
    foo_dir.mkdir(parents=True)
    assert file_utils._is_valid_project_model_path(foo_dir) is False


def test_get_project_model_path(tmp_path: pathlib.Path) -> None:
    """Test get_project_model_path  method."""
    assert file_utils.extract_project_model_path(tmp_path) is None

    test_utils.ensure_trestle_config_dir(tmp_path)
    assert file_utils.extract_project_model_path(tmp_path) is None

    create_sample_catalog_project(tmp_path)

    catalog_dir = tmp_path / 'catalogs'
    assert file_utils.extract_project_model_path(catalog_dir) is None

    mycatalog_dir = catalog_dir / 'mycatalog'
    assert file_utils.extract_project_model_path(mycatalog_dir) == mycatalog_dir

    metadata_dir = mycatalog_dir / 'metadata'
    assert file_utils.extract_project_model_path(metadata_dir) == mycatalog_dir


def test_load_file(tmp_path: pathlib.Path) -> None:
    """Test load file."""
    json_file_path = test_utils.NIST_SAMPLE_CD_JSON
    yaml_file_path = pathlib.Path.joinpath(test_utils.YAML_TEST_DATA_PATH, 'good_component.yaml')

    assert file_utils.load_file(json_file_path) is not None
    assert file_utils.load_file(yaml_file_path) is not None

    try:
        sample_file_path = tmp_path.joinpath('sample.txt')
        with open(sample_file_path, 'w', encoding=const.FILE_ENCODING):
            file_utils.load_file(sample_file_path)
    except TrestleError:
        pass


def test_get_relative_model_type(tmp_path: pathlib.Path) -> None:
    """Test get model type and alias based on filesystem context."""
    import trestle.common.type_utils as cutils
    with pytest.raises(TrestleError):
        ModelUtils.get_relative_model_type(pathlib.Path('invalidpath'))

    with pytest.raises(TrestleError):
        ModelUtils.get_relative_model_type(pathlib.Path('./'))

    catalogs_dir = pathlib.Path('catalogs')
    mycatalog_dir = catalogs_dir / 'mycatalog'
    catalog_dir = mycatalog_dir / 'catalog'
    metadata_dir = catalog_dir / 'metadata'
    roles_dir = metadata_dir / 'roles'
    rps_dir = metadata_dir / 'responsible-parties'
    props_dir = metadata_dir / 'props'
    groups_dir = catalog_dir / 'groups'
    group_dir = groups_dir / f'00000{const.IDX_SEP}group'
    controls_dir = group_dir / 'controls'
    with pytest.raises(TrestleError):
        ModelUtils.get_relative_model_type(catalogs_dir)

    assert ModelUtils.get_relative_model_type(mycatalog_dir) == (catalog.Catalog, 'catalog')
    assert ModelUtils.get_relative_model_type(mycatalog_dir / 'catalog.json') == (catalog.Catalog, 'catalog')
    assert ModelUtils.get_relative_model_type(catalog_dir / 'back-matter.json'
                                              ) == (common.BackMatter, 'catalog.back-matter')
    assert ModelUtils.get_relative_model_type(catalog_dir / 'metadata.yaml') == (common.Metadata, 'catalog.metadata')
    assert ModelUtils.get_relative_model_type(metadata_dir) == (common.Metadata, 'catalog.metadata')
    assert ModelUtils.get_relative_model_type(roles_dir) == (List[common.Role], 'catalog.metadata.roles')
    (type_, element) = ModelUtils.get_relative_model_type(roles_dir)
    assert cutils.get_origin(type_) == list
    assert element == 'catalog.metadata.roles'
    assert ModelUtils.get_relative_model_type(roles_dir / '00000__role.json'
                                              ) == (common.Role, 'catalog.metadata.roles.role')
    model_type, full_alias = ModelUtils.get_relative_model_type(rps_dir)
    assert model_type == List[common.ResponsibleParty]
    assert full_alias == 'catalog.metadata.responsible-parties'
    assert ModelUtils.get_relative_model_type(
        rps_dir / 'creator__responsible-party.json'
    ) == (common.ResponsibleParty, 'catalog.metadata.responsible-parties.responsible-party')
    (type_, element) = ModelUtils.get_relative_model_type(props_dir)
    assert cutils.get_origin(type_) == list
    assert cutils.get_inner_type(type_) == common.Property
    assert element == 'catalog.metadata.props'
    (expected_type, expected_json_path) = ModelUtils.get_relative_model_type(
        props_dir / f'00000{const.IDX_SEP}property.json'
    )
    assert expected_type == common.Property
    assert expected_json_path == 'catalog.metadata.props.property'
    assert cutils.get_origin(type_) == list
    assert ModelUtils.get_relative_model_type(groups_dir / f'00000{const.IDX_SEP}group.json'
                                              ) == (catalog.Group, 'catalog.groups.group')
    assert ModelUtils.get_relative_model_type(group_dir) == (catalog.Group, 'catalog.groups.group')
    assert ModelUtils.get_relative_model_type(controls_dir / f'00000{const.IDX_SEP}control.json'
                                              ) == (catalog.Control, 'catalog.groups.group.controls.control')


def create_sample_catalog_project(trestle_base_dir: pathlib.Path) -> None:
    """Create directory structure for a sample catalog named mycatalog."""
    test_utils.ensure_trestle_config_dir(trestle_base_dir)

    mycatalog_dir = trestle_base_dir / 'catalogs' / 'mycatalog'

    directories = [
        mycatalog_dir / 'catalog' / 'metadata' / 'roles',
        mycatalog_dir / 'catalog' / 'metadata' / 'responsible-parties',
        mycatalog_dir / 'catalog' / 'metadata' / 'props',
        mycatalog_dir / 'catalog' / 'groups' / f'00000{const.IDX_SEP}group' / 'controls'
    ]

    for directory in directories:
        directory.mkdir(parents=True, exist_ok=True)

    files = [
        mycatalog_dir / 'catalog.json',
        mycatalog_dir / 'catalog' / 'back-matter.json',
        mycatalog_dir / 'catalog' / 'metadata.json',
        mycatalog_dir / 'catalog' / 'metadata' / 'roles' / f'00000{const.IDX_SEP}role.json',
        mycatalog_dir / 'catalog' / 'metadata' / 'responsible-parties'
        / f'creator{const.IDX_SEP}responsible-party.json',
        mycatalog_dir / 'catalog' / 'metadata' / 'props' / f'00000{const.IDX_SEP}property.json',
        mycatalog_dir / 'catalog' / 'groups' / f'00000{const.IDX_SEP}group.json',
        mycatalog_dir / 'catalog' / 'groups' / f'00000{const.IDX_SEP}group' / 'controls'
        / f'00000{const.IDX_SEP}control.json',
    ]

    for file in files:
        file.touch()


def test_extract_alias() -> None:
    """Test extraction of alias from filename or directory names."""
    assert ModelUtils._extract_alias(pathlib.Path('catalog').name) == 'catalog'
    assert ModelUtils._extract_alias(pathlib.Path('/tmp/catalog').name) == 'catalog'
    assert ModelUtils._extract_alias(pathlib.Path('/catalogs/mycatalog/catalog.json').name) == 'catalog'
    assert ModelUtils._extract_alias(pathlib.Path('/catalogs/mycatalog/catalog.yaml').name) == 'catalog'
    assert ModelUtils._extract_alias(pathlib.Path('responsible-parties').name) == 'responsible-parties'
    assert ModelUtils._extract_alias(pathlib.Path('responsible-parties.json').name) == 'responsible-parties'
    assert ModelUtils._extract_alias(pathlib.Path('/roles').name) == 'roles'
    assert ModelUtils._extract_alias(pathlib.Path('/roles/roles.json').name) == 'roles'
    assert ModelUtils._extract_alias(pathlib.Path(f'/roles/00000{const.IDX_SEP}role.json').name) == 'role'
    assert ModelUtils._extract_alias(
        pathlib.Path(f'/metadata/responsible-parties/creator{const.IDX_SEP}responsible-party.json').name
    ) == 'responsible-party'


def test_get_stripped_model_type(tmp_path: pathlib.Path) -> None:
    """Test get stripped model type and alias based on filesystem context."""
    with pytest.raises(TrestleError):
        ModelUtils.get_stripped_model_type(tmp_path / 'invalidpath', tmp_path)

    with pytest.raises(TrestleError):
        ModelUtils.get_stripped_model_type(tmp_path, tmp_path)

    create_sample_catalog_project(tmp_path)

    catalogs_dir = tmp_path / 'catalogs'
    with pytest.raises(TrestleError):
        ModelUtils.get_stripped_model_type(catalogs_dir, tmp_path)

    def check_stripped_catalog() -> None:
        assert 'uuid' in alias_to_field_map
        assert 'metadata' not in alias_to_field_map
        assert 'back-matter' not in alias_to_field_map
        assert 'groups' not in alias_to_field_map

    mycatalog_dir = catalogs_dir / 'mycatalog'
    stripped_catalog = ModelUtils.get_stripped_model_type(mycatalog_dir, tmp_path)
    alias_to_field_map = stripped_catalog[0].alias_to_field_map()
    check_stripped_catalog()

    stripped_catalog = ModelUtils.get_stripped_model_type(mycatalog_dir / 'catalog.json', tmp_path)
    alias_to_field_map = stripped_catalog[0].alias_to_field_map()
    check_stripped_catalog()

    def check_stripped_metadata(a2f_map) -> None:
        assert 'title' in a2f_map
        assert 'published' in a2f_map
        assert 'last-modified' in a2f_map
        assert 'version' in a2f_map
        assert 'oscal-version' in a2f_map
        assert 'revisions' in a2f_map
        assert 'document-ids' in a2f_map
        assert 'links' in a2f_map
        assert 'locations' in a2f_map
        assert 'parties' in a2f_map
        assert 'remarks' in a2f_map
        assert 'roles' not in alias_to_field_map
        assert 'responsible-properties' not in a2f_map
        assert 'props' not in a2f_map

    catalog_dir = mycatalog_dir / 'catalog'
    metadata_dir = catalog_dir / 'metadata'
    stripped_catalog = ModelUtils.get_stripped_model_type(metadata_dir, tmp_path)
    alias_to_field_map = stripped_catalog[0].alias_to_field_map()
    check_stripped_metadata(alias_to_field_map)

    stripped_catalog = ModelUtils.get_stripped_model_type(catalog_dir / 'metadata.json', tmp_path)
    alias_to_field_map = stripped_catalog[0].alias_to_field_map()
    check_stripped_metadata(alias_to_field_map)

    groups_dir = catalog_dir / 'groups'
    stripped_catalog = ModelUtils.get_stripped_model_type(groups_dir, tmp_path)

    assert stripped_catalog[0].__name__ == 'Groups'
    assert stripped_catalog[1] == 'catalog.groups'

    def check_stripped_group() -> None:
        assert 'id' in alias_to_field_map
        assert 'class' in alias_to_field_map
        assert 'title' in alias_to_field_map
        assert 'params' in alias_to_field_map
        assert 'props' in alias_to_field_map
        assert 'links' in alias_to_field_map
        assert 'parts' in alias_to_field_map
        assert 'groups' in alias_to_field_map
        assert 'controls' not in alias_to_field_map

    stripped_catalog = ModelUtils.get_stripped_model_type(groups_dir / f'00000{const.IDX_SEP}group', tmp_path)
    alias_to_field_map = stripped_catalog[0].alias_to_field_map()
    check_stripped_group()

    stripped_catalog = ModelUtils.get_stripped_model_type(groups_dir / f'00000{const.IDX_SEP}group.json', tmp_path)
    alias_to_field_map = stripped_catalog[0].alias_to_field_map()
    check_stripped_group()


def test_get_singular_alias() -> None:
    """Test get_singular_alias function."""
    assert ModelUtils.get_singular_alias(alias_path='catalog') == 'catalog'

    # Not fullpath. It should be 'catalog.metadata' instead
    with pytest.raises(TrestleError):
        ModelUtils.get_singular_alias(alias_path='metadata.something')

    # Invalid alias_path
    with pytest.raises(TrestleError):
        ModelUtils.get_singular_alias(alias_path='invalid')
    # Invalid alias_path
    with pytest.raises(TrestleError):
        ModelUtils.get_singular_alias(alias_path='')

    assert ModelUtils.get_singular_alias(alias_path='catalog.metadata.responsible-parties') == 'responsible-party'
    assert ModelUtils.get_singular_alias(alias_path='catalog.metadata.responsible-parties.*.props') == 'property'
    assert 'responsible-party' == ModelUtils.get_singular_alias(alias_path='catalog.metadata.responsible-parties.*')

    assert 'role' == ModelUtils.get_singular_alias(alias_path='catalog.metadata.roles')
    assert 'property' == ModelUtils.get_singular_alias(alias_path='catalog.metadata.props')

    assert 'control-implementations' == ModelUtils.get_singular_alias(
        alias_path='component-definition.components.control-implementations'
    )
    assert 'control-implementation' == ModelUtils.get_singular_alias(
        alias_path='component-definition.components.*.control-implementations'
    )
    assert 'control-implementation' == ModelUtils.get_singular_alias(
        alias_path='component-definition.components.0.control-implementations'
    )
    # FIXME ideally this should report error
    assert '0' == ModelUtils.get_singular_alias(alias_path='component-definition.components.0')

    assert 'control' == ModelUtils.get_singular_alias(alias_path='catalog.groups.*.controls.*.controls')


def test_contextual_get_singular_alias(tmp_path: pathlib.Path, keep_cwd: pathlib.Path) -> None:
    """Test get_singular_alias in contextual mode."""
    # Contextual model tests
    create_sample_catalog_project(tmp_path)
    catalogs_dir = tmp_path.resolve() / 'catalogs'
    mycatalog_dir = catalogs_dir / 'mycatalog'
    catalog_dir = mycatalog_dir / 'catalog'
    metadata_dir = catalog_dir / 'metadata'
    groups_dir = catalog_dir / 'groups'
    group_dir = groups_dir / f'00000{const.IDX_SEP}group'

    rel_dir = mycatalog_dir.relative_to(tmp_path)
    assert 'responsible-party' == ModelUtils.get_singular_alias('catalog.metadata.responsible-parties', rel_dir)
    # Both should work to deal with the case back-matter is already split from the catalog in a separate file
    assert 'resource' == ModelUtils.get_singular_alias('catalog.back-matter.resources', rel_dir)
    assert 'resource' == ModelUtils.get_singular_alias('back-matter.resources', rel_dir)

    rel_dir = metadata_dir.relative_to(tmp_path)
    with pytest.raises(TrestleError):
        ModelUtils.get_singular_alias('metadata.roles')
    alias = ModelUtils.get_singular_alias('metadata.roles', rel_dir)
    assert alias == 'role'
    assert 'responsible-party' == ModelUtils.get_singular_alias('metadata.responsible-parties.*', rel_dir)

    assert 'property' == ModelUtils.get_singular_alias('metadata.responsible-parties.*.props', rel_dir)

    rel_dir = groups_dir.relative_to(tmp_path)
    assert 'control' == ModelUtils.get_singular_alias('groups.*.controls.*.controls', rel_dir)

    rel_dir = group_dir.relative_to(tmp_path)
    assert 'control' == ModelUtils.get_singular_alias('group.controls.*.controls', rel_dir)


def test_get_contextual_file_type(tmp_path: pathlib.Path) -> None:
    """Test fs.get_contextual_file_type()."""
    (tmp_path / 'file.json').touch()
    with pytest.raises(TrestleError):
        file_utils.get_contextual_file_type(pathlib.Path(tmp_path / 'gu.json'))
    (tmp_path / 'file.json').unlink()

    (tmp_path / '.trestle').mkdir()
    (tmp_path / 'catalogs').mkdir()
    catalogs_dir = tmp_path / 'catalogs'
    (catalogs_dir / 'mycatalog').mkdir()
    mycatalog_dir = catalogs_dir / 'mycatalog'

    pathlib.Path(mycatalog_dir / 'file2.json').touch()
    assert file_utils.get_contextual_file_type(mycatalog_dir) == FileContentType.JSON
    (mycatalog_dir / 'file2.json').unlink()

    if file_utils.is_windows():
        hidden_file = mycatalog_dir / 'hidden.txt'
        hidden_file.touch()
        atts = win32api.GetFileAttributes(str(hidden_file))
        win32api.SetFileAttributes(str(hidden_file), win32con.FILE_ATTRIBUTE_HIDDEN | atts)
    else:
        pathlib.Path(mycatalog_dir / '.DS_Store').touch()

    pathlib.Path(mycatalog_dir / 'file2.json').touch()
    assert file_utils.get_contextual_file_type(mycatalog_dir) == FileContentType.JSON

    if file_utils.is_windows():
        hidden_file.unlink()
    else:
        (mycatalog_dir / '.DS_Store').unlink()
    (mycatalog_dir / 'file2.json').unlink()

    pathlib.Path(mycatalog_dir / 'file3.yml').touch()
    assert file_utils.get_contextual_file_type(mycatalog_dir) == FileContentType.YAML
    (mycatalog_dir / 'file3.yml').unlink()

    (mycatalog_dir / 'catalog').mkdir()
    (mycatalog_dir / 'catalog/groups').mkdir()
    (mycatalog_dir / 'catalog/groups/file4.yaml').touch()
    assert file_utils.get_contextual_file_type(mycatalog_dir) == FileContentType.YAML


def test_get_models_of_type(tmp_trestle_dir) -> None:
    """Test fs.get_models_of_type()."""
    create_sample_catalog_project(tmp_trestle_dir)
    catalogs_dir = tmp_trestle_dir.resolve() / 'catalogs'
    components_dir = tmp_trestle_dir.resolve() / 'component-definitions'
    # mycatalog is already there
    (catalogs_dir / 'mycatalog2').mkdir()
    (catalogs_dir / '.myfile').touch()
    (components_dir / 'my_component').mkdir()
    models = ModelUtils.get_models_of_type('catalog', tmp_trestle_dir)
    assert len(models) == 2
    assert 'mycatalog' in models
    assert 'mycatalog2' in models
    all_models = ModelUtils.get_all_models(tmp_trestle_dir)
    assert len(all_models) == 3
    assert ('catalog', 'mycatalog') in all_models
    assert ('catalog', 'mycatalog2') in all_models
    assert ('component-definition', 'my_component') in all_models
    with pytest.raises(TrestleError):
        ModelUtils.get_models_of_type('foo', tmp_trestle_dir)


def test_get_models_of_type_bad_cwd(tmp_path) -> None:
    """Test fs.get_models_of_type() from outside trestle dir."""
    with pytest.raises(TrestleError):
        ModelUtils.get_models_of_type('catalog', tmp_path)


def test_is_hidden_posix(tmp_path) -> None:
    """Test is_hidden on posix systems."""
    if not file_utils.is_windows():
        hidden_file = tmp_path / '.hidden.md'
        hidden_dir = tmp_path / '.hidden/'
        visible_file = tmp_path / 'visible.md'
        visible_dir = tmp_path / 'visible/'

        assert file_utils.is_hidden(hidden_file)
        assert file_utils.is_hidden(hidden_dir)
        assert not file_utils.is_hidden(visible_file)
        assert not file_utils.is_hidden(visible_dir)
    else:
        pass


def test_is_hidden_windows(tmp_path) -> None:
    """Test is_hidden on windows systems."""
    if file_utils.is_windows():
        visible_file = tmp_path / 'visible.md'
        visible_dir = tmp_path / 'visible/'
        visible_file.touch()
        visible_dir.touch()
        assert not file_utils.is_hidden(visible_file)
        assert not file_utils.is_hidden(visible_dir)

        atts = win32api.GetFileAttributes(str(visible_file))
        win32api.SetFileAttributes(str(visible_file), win32con.FILE_ATTRIBUTE_HIDDEN | atts)
        atts = win32api.GetFileAttributes(str(visible_dir))
        win32api.SetFileAttributes(str(visible_dir), win32con.FILE_ATTRIBUTE_HIDDEN | atts)

        assert file_utils.is_hidden(visible_file)
        assert file_utils.is_hidden(visible_dir)
    else:
        pass


@pytest.mark.parametrize(
    'task_name, outcome',
    [
        ('hello', True), ('.trestle', False), ('task/name', True), ('.bad,', False), ('catalogs', False),
        ('catalog', True), ('component-definitions', False), ('hello.world', False),
        ('component-definitions/hello', False)
    ]
)
def test_allowed_task_name(task_name: str, outcome: bool) -> None:
    """Test whether task names are allowed."""
    assert file_utils.is_directory_name_allowed(task_name) == outcome


def test_model_type_to_model_dir() -> None:
    """Test model type to model dir."""
    assert ModelUtils.model_type_to_model_dir('catalog') == 'catalogs'
    try:
        ModelUtils.model_type_to_model_dir('foo')
    except Exception:
        pass
    else:
        assert 'test failed'


def test_local_and_visible(tmp_path) -> None:
    """Test if file is local (not symlink) and visible (not hidden)."""
    local_file = tmp_path / 'local.md'
    local_file.touch()
    if file_utils.is_windows():
        link_file = tmp_path / 'not_local.lnk'
        link_file.touch()
    else:
        link_file = tmp_path / 'linked.md'
        link_file.symlink_to(local_file)
    assert file_utils.is_local_and_visible(local_file)
    assert not file_utils.is_local_and_visible(link_file)


@pytest.mark.parametrize(
    'candidate, build, expect_failure',
    [
        (pathlib.Path('relative_file.json'), False, False),
        (pathlib.Path('relative_file.json'), True, False),
        (pathlib.Path('/random/absolute/path'), False, True),
        (pathlib.Path('/random/absolute/path'), False, True),
        (pathlib.Path('~/random/home_directory/path'), False, True),
        (pathlib.Path('~/random/home_directory/path'), True, False),
        (pathlib.Path('../relative_file.json'), False, True),
        (pathlib.Path('../relative_file.json'), True, True),
        (
            pathlib.Path('./hello/../relative_file.json'),
            False,
            False,
        ),
        (
            pathlib.Path('./hello/../relative_file.json'),
            True,
            False,
        ),
    ]
)
def test_relative_resolve(tmp_path, candidate: pathlib.Path, build: bool, expect_failure: bool):
    """Test relative resolve capability."""
    if build:
        input_path = tmp_path / candidate
    else:
        input_path = candidate
    if expect_failure:
        with pytest.raises(TrestleError):
            _ = file_utils.relative_resolve(input_path, tmp_path)
    else:
        _ = file_utils.relative_resolve(input_path, tmp_path)


def test_iterdir_without_hidden_files(tmp_path: pathlib.Path) -> None:
    """Test that hidden files are filtered from the path."""
    pathlib.Path(tmp_path / 'visible.txt').touch()
    pathlib.Path(tmp_path / 'visibleDir/').mkdir()

    if file_utils.is_windows():
        """Windows"""
        hidden_file = tmp_path / 'hidden.txt'
        hidden_dir = tmp_path / 'hiddenDir/'
        hidden_file.touch()
        hidden_dir.mkdir()
        atts = win32api.GetFileAttributes(str(hidden_file))
        win32api.SetFileAttributes(str(hidden_file), win32con.FILE_ATTRIBUTE_HIDDEN | atts)
        atts = win32api.GetFileAttributes(str(hidden_dir))
        win32api.SetFileAttributes(str(hidden_dir), win32con.FILE_ATTRIBUTE_HIDDEN | atts)

        assert len(list(file_utils.iterdir_without_hidden_files(tmp_path))) == 3
    else:

        pathlib.Path(tmp_path / '.DS_Store').touch()
        pathlib.Path(tmp_path / '.hidden.txt').touch()
        pathlib.Path(tmp_path / '.hiddenDir/').mkdir()

        assert len(list(file_utils.iterdir_without_hidden_files(tmp_path))) == 3


def test_make_hidden_file(tmp_path: pathlib.Path) -> None:
    """Test make hidden files."""
    file_path = tmp_path / '.keep'
    file_utils.make_hidden_file(file_path)

    file_path2 = tmp_path / 'hidden.txt'
    file_utils.make_hidden_file(file_path2)

    assert file_path.exists() and not file_utils.is_local_and_visible(file_path)
    if file_utils.is_windows():
        assert file_path2.exists() and not file_utils.is_local_and_visible(file_path2)
    else:
        assert (tmp_path / '.hidden.txt').exists() and not file_utils.is_local_and_visible(tmp_path / '.hidden.txt')


def test_full_path_for_top_level_model(tmp_trestle_dir: pathlib.Path, sample_catalog_minimal: catalog.Catalog) -> None:
    """Test full path for top level model."""
    ModelUtils.save_top_level_model(sample_catalog_minimal, tmp_trestle_dir, 'mycat', FileContentType.JSON)
    cat_path = ModelUtils.get_model_path_for_name_and_class(tmp_trestle_dir, 'mycat', catalog.Catalog)
    assert cat_path == tmp_trestle_dir / 'catalogs/mycat/catalog.json'


def test_update_last_modified(sample_catalog_rich_controls: catalog.Catalog) -> None:
    """Test update timestamps."""
    hour_ago = datetime.now().astimezone() - timedelta(seconds=const.HOUR_SECONDS)
    ModelUtils.update_last_modified(sample_catalog_rich_controls, hour_ago)
    assert sample_catalog_rich_controls.metadata.last_modified == hour_ago
    ModelUtils.update_last_modified(sample_catalog_rich_controls)
    assert ModelUtils.model_age(sample_catalog_rich_controls) < test_utils.NEW_MODEL_AGE_SECONDS



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the component author module."""

import pathlib
import shutil
from typing import Any, Dict

from _pytest.monkeypatch import MonkeyPatch

from tests import test_utils

import trestle.core.generic_oscal as generic
import trestle.oscal.component as comp
from trestle.common import const, file_utils, model_utils
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.control_interface import ControlInterface
from trestle.core.markdown.markdown_processor import MarkdownProcessor

md_path = 'md_comp'


def add_comp(comp_path: pathlib.Path, ac1_path: pathlib.Path) -> None:
    """Add a new component to the markdown."""
    ac_path = comp_path / 'comp_new/ac'
    ac_path.mkdir(parents=True, exist_ok=True)
    new_ac1_path = ac_path / 'ac-1.md'
    shutil.copyfile(str(ac1_path), str(new_ac1_path))


def check_common_contents(header: Dict[str, Any]) -> None:
    """Check common features of controls markdown."""
    params = header[const.RULES_PARAMS_TAG]['comp_aa']
    assert len(params) == 1
    assert params[0] == {
        'name': 'shared_param_1',
        'description': 'shared param 1 in aa',
        'options': '["shared_param_1_aa_opt_1", "shared_param_1_aa_opt_2", "shared_param_1_aa_opt_3"]',
        'rule-id': 'top_shared_rule_1'
    }
    assert header[const.TRESTLE_GLOBAL_TAG][const.PROFILE]['title'] == 'comp prof aa'
    assert header[const.TRESTLE_GLOBAL_TAG][const.PROFILE
                                            ]['href'] == 'trestle://profiles/comp_prof_aa/profile.json'  # noqa E501


def check_ac1_contents(ac1_path: pathlib.Path) -> None:
    """Check the contents of ac-1 md."""
    assert test_utils.confirm_text_in_file(ac1_path, 'enter one of:', 'ac-1 from comp aa')
    assert test_utils.confirm_text_in_file(ac1_path, 'ac-1 from comp aa', 'Status: implemented')
    assert test_utils.confirm_text_in_file(ac1_path, '- comp_rule_aa_1', 'Status: partial')
    assert test_utils.confirm_text_in_file(ac1_path, 'ac-1_prm_3:', '- set by comp aa imp req')
    markdown_processor = MarkdownProcessor()
    header, _ = markdown_processor.read_markdown_wo_processing(ac1_path)
    assert header[const.PARAM_VALUES_TAG]['ac-1_prm_1'] == ['prof_aa val 1']
    rules = header[const.COMP_DEF_RULES_TAG]['comp_aa']
    assert len(rules) == 2
    assert rules[0] == {'name': 'top_shared_rule_1', 'description': 'top shared rule 1 in aa'}
    assert rules[1] == {'name': 'comp_rule_aa_1', 'description': 'comp rule aa 1'}
    vals = header[const.COMP_DEF_RULES_PARAM_VALS_TAG]['comp_aa']
    assert len(vals) == 1
    assert vals[0]['name'] == 'shared_param_1'
    assert vals[0]['values'] == ['shared_param_1_aa_opt_1']
    check_common_contents(header)


def check_at1_contents(at1_path: pathlib.Path) -> None:
    """Check the contents of at-1 md."""
    markdown_processor = MarkdownProcessor()
    header, _ = markdown_processor.read_markdown_wo_processing(at1_path)
    rules = header[const.COMP_DEF_RULES_TAG]['comp_ab']
    assert len(rules) == 2
    assert rules[0] == {'name': 'XCCDF', 'description': 'The XCCDF must be compliant'}
    assert rules[1] == {'name': 'FancyXtraRule', 'description': 'This is a fancy extra rule'}
    check_common_contents(header)


def test_component_generate(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test component generate."""

    def _assemble_and_check_ac1_contents(assem_name: str, assemble_cmd: str, new_prose: str, add_comp: bool) -> None:
        test_utils.execute_command_and_assert(assemble_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)
        assem_component, _ = model_utils.ModelUtils.load_model_for_class(
            tmp_trestle_dir, assem_name, comp.ComponentDefinition
        )
        for ii in range(3):
            assert assem_component.components[0].control_implementations[0].implemented_requirements[0].set_parameters[
                0].values[ii] == f'inserted value {ii}'

        statement = assem_component.components[0].control_implementations[0].implemented_requirements[0].statements[0]
        assert statement.description == new_prose
        assert ControlInterface.get_status_from_props(statement).state == const.STATUS_IMPLEMENTED
        if add_comp:
            assert assem_component.components[2].title == 'comp_new'

    comp_name = test_utils.setup_component_generate(tmp_trestle_dir)
    ac1_path = tmp_trestle_dir / 'md_comp/comp_aa/comp_prof_aa/ac/ac-1.md'

    orig_component, _ = model_utils.ModelUtils.load_model_for_class(
        tmp_trestle_dir, comp_name, comp.ComponentDefinition
    )

    generate_cmd = f'trestle author component-generate -n {comp_name} -o {md_path}'

    # generate the md first time
    test_utils.execute_command_and_assert(generate_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)
    check_ac1_contents(ac1_path)

    file_checker = test_utils.FileChecker(tmp_trestle_dir / md_path)

    # confirm it overwrites existing md identically
    generate_cmd = f'trestle author component-generate -n {comp_name} -o {md_path}'
    test_utils.execute_command_and_assert(generate_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)

    # all files should be the same
    assert file_checker.files_unchanged()

    assem_name = 'assem_comp'

    # confirm assembled is identical except for uuids
    assemble_cmd = f'trestle author component-assemble -m {md_path} -n {comp_name} -o {assem_name}'
    test_utils.execute_command_and_assert(assemble_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)
    assem_component, assem_comp_path = model_utils.ModelUtils.load_model_for_class(
        tmp_trestle_dir, assem_name, comp.ComponentDefinition
    )
    creation_time = assem_comp_path.stat().st_mtime

    # remove statement ac-1_smt.b because it has no rules associated and won't get written out
    orig_component.components[1].control_implementations[0].implemented_requirements[0].statements.pop()
    assert model_utils.ModelUtils.models_are_equivalent(orig_component, assem_component, True)

    # assemble again and confirm it is not written out since no change
    test_utils.execute_command_and_assert(assemble_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)
    assert creation_time == assem_comp_path.stat().st_mtime

    # edit a rule param value
    new_text = '      component-values:\n        - inserted value 0\n        - inserted value 1\n        - inserted value 2\n'  # noqa E501
    file_utils.insert_text_in_file(ac1_path, '- shared_param_1_aa_opt_1', new_text)

    # edit a status
    test_utils.substitute_text_in_file(
        ac1_path, '### Implementation Status: partial', f'### Implementation Status: {const.STATUS_IMPLEMENTED}'
    )

    # edit prose
    new_prose = 'new prose\nmultiline too'
    test_utils.substitute_text_in_file(ac1_path, 'statement prose for part a. from comp aa', new_prose)

    _assemble_and_check_ac1_contents(assem_name, assemble_cmd, new_prose, False)

    # confirm we can add a new component via markdown
    add_comp(tmp_trestle_dir / 'md_comp', ac1_path)
    _assemble_and_check_ac1_contents(assem_name, assemble_cmd, new_prose, True)


def test_generic_oscal() -> None:
    """Test generic oscal conversions."""
    generic_component = generic.GenericComponent.generate()
    def_comp = generic_component.as_defined_component()
    assert def_comp.description == ''

    generic_cont_imp = generic.GenericControlImplementation.generate()
    cont_imp = generic_cont_imp.as_ssp()
    assert cont_imp.description == ''


def test_component_generate_missing_control(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch, capsys) -> None:
    """Test component generate succeeds when profile missing control."""
    comp_name = test_utils.setup_component_generate(tmp_trestle_dir)

    prof_path = tmp_trestle_dir / 'profiles/comp_prof_aa/profile.json'
    test_utils.delete_line_in_file(prof_path, 'ac-1')

    generate_cmd = f'trestle author component-generate -n {comp_name} -o {md_path}'

    # confirm success when profile is missing a needed control
    test_utils.execute_command_and_assert(generate_cmd, 0, monkeypatch)
    _, err = capsys.readouterr()
    assert "Component comp_aa references controls {'ac-1'} not in profile." in err


def test_component_generate_more_than_one_param(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test component generate with more than 1 parameters per rule."""
    comp_name = test_utils.setup_component_generate(tmp_trestle_dir, 'comp_def_more_params')

    generate_cmd = f'trestle author component-generate -n {comp_name} -o {md_path}'

    # generate the md first time
    test_utils.execute_command_and_assert(generate_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)

    assem_name = 'assem_comp'
    # now assemble component generated
    assemble_cmd = f'trestle author component-assemble -m {md_path} -n {comp_name} -o {assem_name}'
    test_utils.execute_command_and_assert(assemble_cmd, CmdReturnCodes.SUCCESS.value, monkeypatch)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A docs markdown node."""
from __future__ import annotations

import logging
import re
from typing import List, Optional, Tuple

import trestle.core.markdown.markdown_const as md_const
from trestle.core.markdown.base_markdown_node import BaseMarkdownNode, BaseSectionContent

logger = logging.getLogger(__name__)


class DocsSectionContent(BaseSectionContent):
    """A content of the node."""

    def __init__(self):
        """Initialize section content."""
        super(DocsSectionContent, self).__init__()
        self.tables = []
        self.text = []
        self.code_lines = []
        self.html_lines = []
        self.blockquotes = []
        self.governed_document = []

    def union(self, node: DocsMarkdownNode) -> None:
        """Unites contents together."""
        super().union(node)
        self.code_lines.extend(node.content.code_lines)
        self.html_lines.extend(node.content.html_lines)
        self.tables.extend(node.content.tables)
        self.blockquotes.extend(node.content.blockquotes)


class DocsMarkdownNode(BaseMarkdownNode):
    """Markdown will be read to the tree."""

    def __init__(self, key: str, content: DocsSectionContent, starting_line: int):
        """Initialize markdown node."""
        super(DocsMarkdownNode, self).__init__(key, content, starting_line)
        self.content: DocsSectionContent = content

    @classmethod
    def build_tree_from_markdown(cls, lines: List[str], governed_header: Optional[str] = None):
        """Construct a tree out of the given markdown."""
        ob = cls.__new__(cls)
        start_level = ob._get_max_header_lvl(lines)
        ob, _ = ob._build_tree(lines, 'root', 0, start_level, governed_header)
        return ob

    def _build_tree(
        self,
        lines: List[str],
        root_key: str,
        starting_line: int,
        level: int,
        governed_header: Optional[str] = None
    ) -> Tuple[DocsMarkdownNode, int]:
        """
        Build a tree from the markdown recursively.

        The tree is contructed with valid headers as node's keys
        and node's content contains everything that is under that header.
        The subsections are placed into node's children with the same structure.

        A header is valid iff the line starts with # and it is not:
          1. Inside of the html blocks
          2. Inside single lined in the <> tags
          3. Inside the html comment
          4. Inside any table, code block or blockquotes
        """
        content = DocsSectionContent()
        node_children = []
        i = starting_line

        while True:
            if i >= len(lines):
                break
            line = lines[i].strip(' ')
            header_lvl = self._get_header_level_if_valid(line)

            if header_lvl is not None:
                if header_lvl >= level + 1:
                    # build subtree
                    subtree, i = self._build_tree(lines, line, i + 1, level + 1, governed_header)
                    node_children.append(subtree)
                    content.union(subtree)
                else:
                    break  # level of the header is above or equal to the current level, subtree is over
            elif self._does_start_with(line, md_const.CODEBLOCK_DEF):
                code_lines, i = self._read_code_lines(lines, line, i + 1)
                content.code_lines.extend(code_lines)
            elif self._does_start_with(line, md_const.HTML_COMMENT_START):
                html_lines, i = self._read_html_block(lines, line, i + 1, md_const.HTML_COMMENT_END_REGEX)
                content.html_lines.extend(html_lines)
            elif self._does_contain(line, md_const.HTML_TAG_REGEX_START):
                html_lines, i = self._read_html_block(lines, line, i + 1, md_const.HTML_TAG_REGEX_END)
                content.html_lines.extend(html_lines)
            elif self._does_start_with(line, md_const.TABLE_SYMBOL):
                table_block, i = self._read_table_block(lines, line, i + 1)
                content.tables.extend(table_block)
            elif self._does_start_with(line, md_const.BLOCKQUOTE_CHAR):
                content.blockquotes.append(line)
                i += 1
            elif governed_header is not None and self._does_contain(
                    root_key, fr'^[#]+ {governed_header}$') and self._does_contain(line, md_const.GOVERNED_DOC_REGEX):
                regexp = re.compile(md_const.GOVERNED_DOC_REGEX)
                match = regexp.search(line)
                header = match.group(0).strip('*').strip(':')
                content.governed_document.append(header)
                i += 1
            else:
                content.text.append(line)
                i += 1

        first_line_to_grab = starting_line - 1 if starting_line else 0
        content.raw_text = '\n'.join(lines[first_line_to_grab:i])
        md_node = DocsMarkdownNode(key=root_key, content=content, starting_line=first_line_to_grab)
        md_node.subnodes = node_children
        return md_node, i



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""oscal-catalog-to-csv task tests."""

import configparser
import csv
import pathlib
from typing import Dict, List

from _pytest.monkeypatch import MonkeyPatch

from tests import test_utils

import trestle.tasks.oscal_catalog_to_csv as oscal_catalog_to_csv
from trestle.core.catalog.catalog_interface import CatalogInterface
from trestle.oscal.common import HowMany
from trestle.tasks.base_task import TaskOutcome

CONFIG_BY_CONTROL = 'test-oscal-catalog-to-csv-rev-5-by-control.config'
CONFIG_BY_STATEMENT = 'test-oscal-catalog-to-csv-rev-5-by-statement.config'

CONFIG_LIST = [f'{CONFIG_BY_CONTROL}', f'{CONFIG_BY_STATEMENT}']


def monkey_exception():
    """Monkey exception."""
    raise RuntimeError('foobar')


def monkey_get_dependent_control_ids(self, control_id: str):
    """Monkey get_dependent_control_ids."""
    return ['parent', 'parent']


def _get_rows(csv_path: pathlib.Path) -> List[List[str]]:
    """Get rows from csv file."""
    rows = []
    with open(csv_path, 'r', newline='') as f:
        csv_reader = csv.reader(f, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        for row in csv_reader:
            rows.append(row)
    return rows


def _validate(config: str, section: Dict[str, str]) -> None:
    """Validate."""
    odir = section['output-dir']
    oname = section['output-name']
    opth = pathlib.Path(odir) / oname
    rows = _get_rows(opth)
    # spot check
    if config == CONFIG_BY_CONTROL:
        assert len(rows) == 1190
        row = rows[0]
        assert row[0] == 'Control Identifier'
        assert row[1] == 'Control Title'
        assert row[2] == 'Control Text'
        row = rows[1]
        assert row[0] == 'AC-1'
        assert row[1] == 'Policy and Procedures'
        assert row[
            2
        ] == 'a. Develop, document, and disseminate to [Assignment: organization-defined personnel or roles]: 1. [Selection (one or more): organization-level; mission/business process-level; system-level] access control policy that: 2. Procedures to facilitate the implementation of the access control policy and the associated access controls; b. Designate an [Assignment: official] to manage the development, documentation, and dissemination of the access control policy and procedures; and c. Review and update the current access control: 1. Policy [Assignment: frequency] and following [Assignment: events] ; and 2. Procedures [Assignment: frequency] and following [Assignment: events].'  # noqa
    elif config == CONFIG_BY_STATEMENT:
        assert len(rows) == 1750
        row = rows[0]
        assert row[0] == 'Control Identifier'
        assert row[1] == 'Control Title'
        assert row[2] == 'Statement Identifier'
        assert row[3] == 'Statement Text'
        row = rows[1]
        assert row[0] == 'AC-1'
        assert row[1] == 'Policy and Procedures'
        assert row[2] == 'AC-1(a)'
        assert row[
            3
        ] == 'a. Develop, document, and disseminate to [Assignment: organization-defined personnel or roles]: 1. [Selection (one or more): organization-level; mission/business process-level; system-level] access control policy that: 2. Procedures to facilitate the implementation of the access control policy and the associated access controls;'  # noqa


def _test_init(tmp_path: pathlib.Path):
    """Test init."""
    test_utils.ensure_trestle_config_dir(tmp_path)


def _get_config_section(tmp_path: pathlib.Path, fname: str) -> tuple:
    """Get config section."""
    config = configparser.ConfigParser()
    config_path = pathlib.Path(f'tests/data/tasks/oscal-catalog-to-csv/{fname}')
    config.read(config_path)
    section = config['task.oscal-catalog-to-csv']
    section['output-dir'] = str(tmp_path)
    return (config, section)


def _get_config_section_init(tmp_path: pathlib.Path, fname: str) -> tuple:
    """Get config section."""
    _test_init(tmp_path)
    return _get_config_section(tmp_path, fname)


def test_print_info(tmp_path: pathlib.Path) -> None:
    """Test print_info."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.print_info()
        assert retval is None


def test_missing_section(tmp_path: pathlib.Path):
    """Test missing section."""
    section = None
    tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
    retval = tgt.execute()
    assert retval == TaskOutcome.FAILURE


def test_missing_input(tmp_path: pathlib.Path):
    """Test missing input."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        section.pop('input-file')
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_missing_output(tmp_path: pathlib.Path):
    """Test missing output."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        section.pop('output-dir')
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_bogus_level(tmp_path: pathlib.Path):
    """Test bogus level."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        section['level'] = 'foobar'
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_simulate(tmp_path: pathlib.Path):
    """Test execute."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        section['output-dir'] = str(tmp_path)
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.simulate()
        assert retval == TaskOutcome.SIM_SUCCESS


def test_execute(tmp_path: pathlib.Path):
    """Test execute."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        section['output-dir'] = str(tmp_path)
        section['output-name'] = f'{config}.csv'
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
        _validate(config, section)


def test_no_overwrite(tmp_path: pathlib.Path):
    """Test no overwrite."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        section['output-dir'] = str(tmp_path)
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.SUCCESS
        section['output-overwrite'] = 'false'
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_exception(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test exception."""
    monkeypatch.setattr(oscal_catalog_to_csv.CsvHelper, 'write', monkey_exception)
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        section['output-dir'] = str(tmp_path)
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE


def test_join():
    """Test join."""
    oscal_catalog_to_csv.join_str(None, None)
    oscal_catalog_to_csv.join_str('x', None)
    oscal_catalog_to_csv.join_str(None, 'y')
    oscal_catalog_to_csv.join_str('x', 'y')


def test_derive_id(tmp_path: pathlib.Path):
    """Test derive_id."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        ifile = section['input-file']
        ipth = pathlib.Path(ifile)
        catalog_helper = oscal_catalog_to_csv.CatalogHelper(ipth)
        ids = ['ac-4.1_smt', 'ac-4.1_smt.a', 'ac-4.1_smt.a.b', 'ac-4.1_smt.a.b.c', 'ac-4.1_smt.a.b.c.d']
        for id_ in ids:
            catalog_helper._derive_id(id_)


def test_unresolved_param(tmp_path: pathlib.Path):
    """Test unresolved param."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        ifile = section['input-file']
        ipth = pathlib.Path(ifile)
        catalog_helper = oscal_catalog_to_csv.CatalogHelper(ipth)
        for control in catalog_helper.get_controls():
            utext = 'foo {{ insert: param, xx-11_odp.02 }} bar'
            try:
                catalog_helper._resolve_parms(control, utext)
                raise RuntimeError('huh?')
            except RuntimeError:
                break


def test_one_choice(tmp_path: pathlib.Path):
    """Test one choice."""
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        ifile = section['input-file']
        ipth = pathlib.Path(ifile)
        catalog_helper = oscal_catalog_to_csv.CatalogHelper(ipth)
        for control in catalog_helper.get_controls():
            if control.params:
                for param in control.params:
                    if param.select:
                        param.select.how_many = HowMany.one
                        catalog_helper._get_parm_value(control, param.id)
                        return


def test_duplicate(tmp_path: pathlib.Path, monkeypatch: MonkeyPatch):
    """Test duplicate."""
    monkeypatch.setattr(CatalogInterface, 'get_dependent_control_ids', monkey_get_dependent_control_ids)
    for config in CONFIG_LIST:
        _, section = _get_config_section_init(tmp_path, config)
        section['output-dir'] = str(tmp_path)
        tgt = oscal_catalog_to_csv.OscalCatalogToCsv(section)
        retval = tgt.execute()
        assert retval == TaskOutcome.FAILURE



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
#
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#                                                                        #
#        This file is automatically generated by gen_oscal.py            #
#                                                                        #
####### DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT DO NOT EDIT ######
#
#
from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import AnyUrl, EmailStr, Extra, Field, conint, constr, validator

from trestle.core.base_model import OscalBaseModel
from trestle.oscal import OSCAL_VERSION_REGEX, OSCAL_VERSION
import trestle.oscal.common as common


class TermsAndConditions(OscalBaseModel):
    """
    Used to define various terms and conditions under which an assessment, described by the plan, can be performed. Each child part defines a different type of term or condition.
    """

    class Config:
        extra = Extra.forbid

    parts: Optional[List[common.AssessmentPart]] = Field(None)


class State1(Enum):
    """
    The operational status.
    """

    under_development = 'under-development'
    operational = 'operational'
    disposition = 'disposition'
    other = 'other'


class State(Enum):
    """
    An indication as to whether the objective is satisfied or not.
    """

    satisfied = 'satisfied'
    not_satisfied = 'not-satisfied'


class Status(OscalBaseModel):
    """
    A determination of if the objective is satisfied or not within a given system.
    """

    class Config:
        extra = Extra.forbid

    state: State = Field(
        ...,
        description='An indication as to whether the objective is satisfied or not.',
        title='Objective Status State',
    )
    reason: Optional[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )] = Field(
        None,
        description="The reason the objective was given it's status.",
        title='Objective Status Reason',
    )
    remarks: Optional[str] = None


class SelectControlById(OscalBaseModel):
    """
    Used to select a control for inclusion/exclusion based on one or more control identifiers. A set of statement identifiers can be used to target the inclusion/exclusion to only specific control statements providing more granularity over the specific statements that are within the asessment scope.
    """

    class Config:
        extra = Extra.forbid

    control_id: constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    ) = Field(
        ...,
        alias='control-id',
        description=
        'A human-oriented identifier reference to a control with a corresponding id value. When referencing an externally defined control, the Control Identifier Reference must be used in the context of the external / imported OSCAL instance (e.g., uri-reference).',
        title='Control Identifier Reference',
    )
    statement_ids: Optional[List[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )]] = Field(
        None, alias='statement-ids'
    )


class RelatedObservation(OscalBaseModel):
    """
    Relates the finding to a set of referenced observations that were used to determine the finding.
    """

    class Config:
        extra = Extra.forbid

    observation_uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        alias='observation-uuid',
        description='A machine-oriented identifier reference to an observation defined in the list of observations.',
        title='Observation Universally Unique Identifier Reference',
    )


class Origin(OscalBaseModel):
    """
    Identifies the source of the finding, such as a tool, interviewed person, or activity.
    """

    class Config:
        extra = Extra.forbid

    actors: List[common.OriginActor] = Field(...)
    related_tasks: Optional[List[common.RelatedTask]] = Field(None, alias='related-tasks')


class Method(OscalBaseModel):
    __root__: constr(regex=r'^\S(.*\S)?$') = Field(
        ...,
        description='Identifies how the observation was made.',
        title='Observation Method',
    )


class Observation(OscalBaseModel):
    """
    Describes an individual observation.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this observation elsewhere in this or other OSCAL instances. The locally defined UUID of the observation can be used to reference the data item locally or globally (e.g., in an imorted OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Observation Universally Unique Identifier',
    )
    title: Optional[str] = Field(None, description='The title for this observation.', title='Observation Title')
    description: str = Field(
        ...,
        description='A human-readable description of this assessment observation.',
        title='Observation Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    methods: List[Method] = Field(...)
    types: Optional[List[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )]] = Field(None)
    origins: Optional[List[Origin]] = Field(None)
    subjects: Optional[List[common.SubjectReference]] = Field(None)
    relevant_evidence: Optional[List[common.RelevantEvidence]] = Field(None, alias='relevant-evidence')
    collected: datetime = Field(
        ...,
        description='Date/time stamp identifying when the finding information was collected.',
        title='Collected Field',
    )
    expires: Optional[datetime] = Field(
        None,
        description=
        'Date/time identifying when the finding information is out-of-date and no longer valid. Typically used with continuous assessment scenarios.',
        title='Expires Field',
    )
    remarks: Optional[str] = None


class Entry(OscalBaseModel):
    """
    Identifies an individual risk response that occurred as part of managing an identified risk.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this risk log entry elsewhere in this or other OSCAL instances. The locally defined UUID of the risk log entry can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Risk Log Entry Universally Unique Identifier',
    )
    title: Optional[str] = Field(None, description='The title for this risk log entry.', title='Title')
    description: Optional[str] = Field(
        None,
        description='A human-readable description of what was done regarding the risk.',
        title='Risk Task Description',
    )
    start: datetime = Field(
        ...,
        description='Identifies the start date and time of the event.',
        title='Start',
    )
    end: Optional[datetime] = Field(
        None,
        description=
        'Identifies the end date and time of the event. If the event is a point in time, the start and end will be the same date and time.',
        title='End',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    logged_by: Optional[List[common.LoggedBy]] = Field(None, alias='logged-by')
    status_change: Optional[constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )] = Field(
        None, alias='status-change'
    )
    related_responses: Optional[List[common.RelatedResponse]] = Field(None, alias='related-responses')
    remarks: Optional[str] = None


class ControlSelection(OscalBaseModel):
    """
    Identifies the controls being assessed. In the assessment plan, these are the planned controls. In the assessment results, these are the actual controls, and reflects any changes from the plan.
    """

    class Config:
        extra = Extra.forbid

    description: Optional[str] = Field(
        None,
        description='A human-readable description of in-scope controls specified for assessment.',
        title='Assessed Controls Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    include_all: Optional[common.IncludeAll] = Field(None, alias='include-all')
    include_controls: Optional[List[SelectControlById]] = Field(None, alias='include-controls')
    exclude_controls: Optional[List[SelectControlById]] = Field(None, alias='exclude-controls')
    remarks: Optional[str] = None


class Characterization(OscalBaseModel):
    """
    A collection of descriptive data about the containing object from a specific origin.
    """

    class Config:
        extra = Extra.forbid

    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    origin: Origin
    facets: List[common.Facet] = Field(...)


class Status1(OscalBaseModel):
    """
    Describes the operational status of the system component.
    """

    class Config:
        extra = Extra.forbid

    state: State1 = Field(..., description='The operational status.', title='State')
    remarks: Optional[str] = None


class SystemComponent(OscalBaseModel):
    """
    A defined component that can be part of an implemented system.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this component elsewhere in this or other OSCAL instances. The locally defined UUID of the component can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Component Identifier',
    )
    type: constr(regex=r'^\S(.*\S)?$') = Field(
        ...,
        description='A category describing the purpose of the component.',
        title='Component Type',
    )
    title: str = Field(
        ...,
        description='A human readable name for the system component.',
        title='Component Title',
    )
    description: str = Field(
        ...,
        description='A description of the component, including information about its function.',
        title='Component Description',
    )
    purpose: Optional[str] = Field(
        None,
        description='A summary of the technological or business purpose of the component.',
        title='Purpose',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    status: Status1 = Field(
        ...,
        description='Describes the operational status of the system component.',
        title='Status',
    )
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    protocols: Optional[List[common.Protocol]] = Field(None)
    remarks: Optional[str] = None


class RiskLog(OscalBaseModel):
    """
    A log of all risk-related tasks taken.
    """

    class Config:
        extra = Extra.forbid

    entries: List[Entry] = Field(...)


class ReviewedControls(OscalBaseModel):
    """
    Identifies the controls being assessed and their control objectives.
    """

    class Config:
        extra = Extra.forbid

    description: Optional[str] = Field(
        None,
        description='A human-readable description of control objectives.',
        title='Control Objective Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    control_selections: List[ControlSelection] = Field(..., alias='control-selections')
    control_objective_selections: Optional[List[common.ControlObjectiveSelection]] = Field(
        None, alias='control-objective-selections'
    )
    remarks: Optional[str] = None


class Response(OscalBaseModel):
    """
    Describes either recommended or an actual plan for addressing the risk.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this remediation elsewhere in this or other OSCAL instances. The locally defined UUID of the risk response can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Remediation Universally Unique Identifier',
    )
    lifecycle: constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    ) = Field(
        ...,
        description=
        'Identifies whether this is a recommendation, such as from an assessor or tool, or an actual plan accepted by the system owner.',
        title='Remediation Intent',
    )
    title: str = Field(..., description='The title for this response activity.', title='Response Title')
    description: str = Field(
        ...,
        description='A human-readable description of this response plan.',
        title='Response Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    origins: Optional[List[Origin]] = Field(None)
    required_assets: Optional[List[common.RequiredAsset]] = Field(None, alias='required-assets')
    tasks: Optional[List[common.Task]] = Field(None)
    remarks: Optional[str] = None


class Risk(OscalBaseModel):
    """
    An identified risk.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this risk elsewhere in this or other OSCAL instances. The locally defined UUID of the risk can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Risk Universally Unique Identifier',
    )
    title: str = Field(..., description='The title for this risk.', title='Risk Title')
    description: str = Field(
        ...,
        description=
        'A human-readable summary of the identified risk, to include a statement of how the risk impacts the system.',
        title='Risk Description',
    )
    statement: str = Field(
        ...,
        description='An summary of impact for how the risk affects the system.',
        title='Risk Statement',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    status: constr(
        regex=
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'
    )
    origins: Optional[List[Origin]] = Field(None)
    threat_ids: Optional[List[common.ThreatId]] = Field(None, alias='threat-ids')
    characterizations: Optional[List[Characterization]] = Field(None)
    mitigating_factors: Optional[List[common.MitigatingFactor]] = Field(None, alias='mitigating-factors')
    deadline: Optional[datetime] = Field(
        None,
        description='The date/time by which the risk must be resolved.',
        title='Risk Resolution Deadline',
    )
    remediations: Optional[List[Response]] = Field(None)
    risk_log: Optional[RiskLog] = Field(
        None,
        alias='risk-log',
        description='A log of all risk-related tasks taken.',
        title='Risk Log',
    )
    related_observations: Optional[List[RelatedObservation]] = Field(None, alias='related-observations')


class AssessmentAssets(OscalBaseModel):
    """
    Identifies the assets used to perform this assessment, such as the assessment team, scanning tools, and assumptions.
    """

    class Config:
        extra = Extra.forbid

    components: Optional[List[SystemComponent]] = Field(None)
    assessment_platforms: List[common.AssessmentPlatform] = Field(..., alias='assessment-platforms')


class Step(OscalBaseModel):
    """
    Identifies an individual step in a series of steps related to an activity, such as an assessment test or examination procedure.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this step elsewhere in this or other OSCAL instances. The locally defined UUID of the step (in a series of steps) can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Step Universally Unique Identifier',
    )
    title: Optional[str] = Field(None, description='The title for this step.', title='Step Title')
    description: str = Field(
        ...,
        description='A human-readable description of this step.',
        title='Step Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    reviewed_controls: Optional[ReviewedControls] = Field(None, alias='reviewed-controls')
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    remarks: Optional[str] = None


class Activity(OscalBaseModel):
    """
    Identifies an assessment or related process that can be performed. In the assessment plan, this is an intended activity which may be associated with an assessment task. In the assessment results, this an activity that was actually performed as part of an assessment.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this assessment activity elsewhere in this or other OSCAL instances. The locally defined UUID of the activity can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Assessment Activity Universally Unique Identifier',
    )
    title: Optional[str] = Field(
        None,
        description='The title for this included activity.',
        title='Included Activity Title',
    )
    description: str = Field(
        ...,
        description='A human-readable description of this included activity.',
        title='Included Activity Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    steps: Optional[List[Step]] = Field(None)
    related_controls: Optional[ReviewedControls] = Field(None, alias='related-controls')
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    remarks: Optional[str] = None


class LocalDefinitions(OscalBaseModel):
    """
    Used to define data objects that are used in the assessment plan, that do not appear in the referenced SSP.
    """

    class Config:
        extra = Extra.forbid

    components: Optional[List[SystemComponent]] = Field(None)
    inventory_items: Optional[List[common.InventoryItem]] = Field(None, alias='inventory-items')
    users: Optional[List[common.SystemUser]] = Field(None)
    objectives_and_methods: Optional[List[common.LocalObjective]] = Field(None, alias='objectives-and-methods')
    activities: Optional[List[Activity]] = Field(None)
    remarks: Optional[str] = None


class AssessmentPlan(OscalBaseModel):
    """
    An assessment plan, such as those provided by a FedRAMP assessor.
    """

    class Config:
        extra = Extra.forbid

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[45][0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'
    ) = Field(
        ...,
        description=
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this assessment plan in this or other OSCAL instances. The locally defined UUID of the assessment plan can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',
        title='Assessment Plan Universally Unique Identifier',
    )
    metadata: common.Metadata
    import_ssp: common.ImportSsp = Field(..., alias='import-ssp')
    local_definitions: Optional[LocalDefinitions] = Field(
        None,
        alias='local-definitions',
        description=
        'Used to define data objects that are used in the assessment plan, that do not appear in the referenced SSP.',
        title='Local Definitions',
    )
    terms_and_conditions: Optional[TermsAndConditions] = Field(
        None,
        alias='terms-and-conditions',
        description=
        'Used to define various terms and conditions under which an assessment, described by the plan, can be performed. Each child part defines a different type of term or condition.',
        title='Assessment Plan Terms and Conditions',
    )
    reviewed_controls: ReviewedControls = Field(..., alias='reviewed-controls')
    assessment_subjects: Optional[List[common.AssessmentSubject]] = Field(None, alias='assessment-subjects')
    assessment_assets: Optional[AssessmentAssets] = Field(None, alias='assessment-assets')
    tasks: Optional[List[common.Task]] = Field(None)
    back_matter: Optional[common.BackMatter] = Field(None, alias='back-matter')


class Model(OscalBaseModel):
    assessment_plan: AssessmentPlan = Field(..., alias='assessment-plan')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle utilities to customize ."""
import logging
from datetime import date

import frontmatter

from jinja2 import lexer, nodes
from jinja2.environment import Environment
from jinja2.ext import Extension
from jinja2.parser import Parser

from trestle.common import err
from trestle.core.markdown import docs_markdown_node, markdown_const

logger = logging.getLogger(__name__)


def adjust_heading_level(input_md: str, expected: int) -> str:
    """Adjust the header level of a markdown string such that the most significant header matches the expected #'s."""
    output_md = input_md
    mdn = docs_markdown_node.DocsMarkdownNode.build_tree_from_markdown(input_md.split('\n'))
    if mdn.subnodes:
        mdn_top_heading = mdn.subnodes[0].get_node_header_lvl()
        delta = int(expected) - mdn_top_heading
        if not delta == 0:
            mdn.change_header_level_by(delta)
            output_md = mdn.content.raw_text
    return output_md


class TrestleJinjaExtension(Extension):
    """Class to define common methods to be inherited from for use in trestle."""

    # This
    max_tag_parse = 20

    def __init__(self, environment: Environment) -> None:
        """Ensure enviroment is set and carried into class vars."""
        super().__init__(environment)

    @staticmethod
    def parse_expression(parser):
        """Safely parse jinja expression."""
        # Licensed under MIT from:
        # https://github.com/MoritzS/jinja2-django-tags/blob/master/jdj_tags/extensions.py#L424
        # Due to how the jinja2 parser works, it treats "foo" "bar" as a single
        # string literal as it is the case in python.
        # But the url tag in django supports multiple string arguments, e.g.
        # "{% url 'my_view' 'arg1' 'arg2' %}".
        # That's why we have to check if it's a string literal first.
        token = parser.stream.current
        if token.test(lexer.TOKEN_STRING):
            expr = nodes.Const(token.value, lineno=token.lineno)
            next(parser.stream)
        else:
            expr = parser.parse_expression(False)

        return expr


class MDSectionInclude(TrestleJinjaExtension):
    """Inject the parameter of the tag as the resulting content."""

    tags = {'mdsection_include'}

    def __init__(self, environment: Environment) -> None:
        """Ensure enviroment is set and carried into class vars."""
        super().__init__(environment)

    def parse(self, parser):
        """Execute parsing of md token and return nodes."""
        kwargs = None
        expected_heading_level = None
        count = 0
        while parser.stream.current.type != lexer.TOKEN_BLOCK_END:
            count = count + 1
            if count > self.max_tag_parse:
                raise err.TrestleError('Unexpected Jinja tag structure provided, please review docs.')
            token = parser.stream.current
            if token.test('name:mdsection_include'):
                parser.stream.expect(lexer.TOKEN_NAME)
                markdown_source = parser.stream.expect(lexer.TOKEN_STRING)
                section_title = parser.stream.expect(lexer.TOKEN_STRING)
            elif kwargs is not None:
                arg = token.value
                next(parser.stream)
                parser.stream.expect(lexer.TOKEN_ASSIGN)
                token = parser.stream.current
                exp = self.parse_expression(parser)
                kwargs[arg] = exp.value
            else:
                if parser.stream.look().type == lexer.TOKEN_ASSIGN:
                    kwargs = {}
                continue
        # Use the established environment to source the file
        md_content, _, _ = self.environment.loader.get_source(self.environment, markdown_source.value)
        fm = frontmatter.loads(md_content)
        if not fm.metadata == {}:
            logger.warning('Non zero metadata on MD section include - ignoring')
        full_md = docs_markdown_node.DocsMarkdownNode.build_tree_from_markdown(fm.content.split('\n'))
        md_section = full_md.get_node_for_key(section_title.value, strict_matching=True)
        # adjust
        if kwargs is not None:
            expected_heading_level = kwargs.get('heading_level')
        if expected_heading_level is not None:
            level = md_section.get_node_header_lvl()
            delta = int(expected_heading_level) - level
            if not delta == 0:
                md_section.change_header_level_by(delta)
        if not md_section:
            raise err.TrestleError(
                f'Unable to retrieve section "{section_title.value}"" from {markdown_source.value} jinja template.'
            )
        local_parser = Parser(self.environment, md_section.content.raw_text)
        top_level_output = local_parser.parse()

        return top_level_output.body


class MDCleanInclude(TrestleJinjaExtension):
    """Inject the parameter of the tag as the resulting content."""

    tags = {'md_clean_include'}

    def __init__(self, environment: Environment) -> None:
        """Ensure enviroment is set and carried into class vars."""
        super().__init__(environment)

    def parse(self, parser):
        """Execute parsing of md token and return nodes."""
        kwargs = None
        expected_heading_level = None
        count = 0
        while parser.stream.current.type != lexer.TOKEN_BLOCK_END:
            count = count + 1
            if count > self.max_tag_parse:
                raise err.TrestleError('Unexpected Jinja tag structure provided, please review docs.')
            token = parser.stream.current
            if token.test('name:md_clean_include'):
                parser.stream.expect(lexer.TOKEN_NAME)
                markdown_source = parser.stream.expect(lexer.TOKEN_STRING)
            elif kwargs is not None:
                arg = token.value
                next(parser.stream)
                parser.stream.expect(lexer.TOKEN_ASSIGN)
                token = parser.stream.current
                exp = self.parse_expression(parser)
                kwargs[arg] = exp.value
            else:
                if parser.stream.look().type == lexer.TOKEN_ASSIGN:
                    kwargs = {}
                continue
        md_content, _, _ = self.environment.loader.get_source(self.environment, markdown_source.value)
        fm = frontmatter.loads(md_content)
        content = fm.content
        content += '\n\n'
        if kwargs is not None:
            expected_heading_level = kwargs.get('heading_level')
        if expected_heading_level is not None:
            content = adjust_heading_level(content, expected_heading_level)

        local_parser = Parser(self.environment, content)
        top_level_output = local_parser.parse()

        return top_level_output.body


class MDDatestamp(TrestleJinjaExtension):
    """Inject the parameter of the tag as the resulting content."""

    tags = {'md_datestamp'}

    def __init__(self, environment: Environment) -> None:
        """Ensure enviroment is set and carried into class vars."""
        super().__init__(environment)

    def parse(self, parser):
        """Execute parsing of md token and return nodes."""
        kwargs = None
        count = 0
        while parser.stream.current.type != lexer.TOKEN_BLOCK_END:
            count = count + 1
            token = parser.stream.current
            if count > self.max_tag_parse:
                raise err.TrestleError(f'Unexpected Jinja tag structure provided at token {token.value}')
            if token.test('name:md_datestamp'):
                parser.stream.expect(lexer.TOKEN_NAME)
            elif kwargs is not None:
                arg = token.value
                next(parser.stream)
                parser.stream.expect(lexer.TOKEN_ASSIGN)
                token = parser.stream.current
                exp = self.parse_expression(parser)
                kwargs[arg] = exp.value
            else:
                if parser.stream.look().type == lexer.TOKEN_ASSIGN or parser.stream.look().type == lexer.TOKEN_STRING:
                    kwargs = {}
                continue

        if kwargs is not None:
            if 'format' in kwargs and type(kwargs['format'] is str):
                date_string = date.today().strftime(kwargs['format'])
            else:
                date_string = date.today().strftime(markdown_const.JINJA_DATESTAMP_FORMAT)
            if 'newline' in kwargs and kwargs['newline'] is False:
                pass
            else:
                date_string += '\n\n'
        else:
            date_string = date.today().strftime(markdown_const.JINJA_DATESTAMP_FORMAT) + '\n\n'

        local_parser = Parser(self.environment, date_string)
        datestamp_output = local_parser.parse()

        return datestamp_output.body



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle replicate command."""
import argparse
import shutil
import sys
from pathlib import Path

from _pytest.monkeypatch import MonkeyPatch

import pytest

from tests import test_utils

import trestle.common.err as err
from trestle.cli import Trestle
from trestle.common import const
from trestle.common.model_utils import ModelUtils
from trestle.core.commands.replicate import ReplicateCmd
from trestle.oscal.catalog import Catalog

subcommand_list = const.MODEL_TYPE_LIST


def mock_return(*args, **kwargs):
    """Mock successful function call with no return value."""
    return


def _copy_local(source_model_path, source_model_name, model_alias):
    plural_alias = ModelUtils.model_type_to_model_dir(model_alias)
    full_model_path = source_model_path / plural_alias / source_model_name
    local_model_path = Path(plural_alias) / source_model_name
    shutil.rmtree(local_model_path, ignore_errors=True)
    shutil.copytree(full_model_path, local_model_path)


def test_replicate_cmd_no_file(tmp_trestle_dir: Path, monkeypatch: MonkeyPatch) -> None:
    """Confirm failure when file does not exist."""
    for subcommand in subcommand_list:
        test_args = f'trestle replicate {subcommand} -n mock_file_name -o random_named_{subcommand}'.split()
        monkeypatch.setattr(sys, 'argv', test_args)
        rc = Trestle().run()
        assert rc != 0


@pytest.mark.parametrize('regen', ['', '-r'])
def test_replicate_cmd(testdata_dir, tmp_trestle_dir, regen, monkeypatch: MonkeyPatch) -> None:
    """Test replicate command."""
    # prepare trestle project dir with the file
    test_utils.ensure_trestle_config_dir(tmp_trestle_dir)

    test_data_source = testdata_dir / 'split_merge/step4_split_groups_array'

    source_name = 'mycatalog'
    rep_name = 'repcatalog'
    _copy_local(test_data_source, source_name, 'catalog')

    catalogs_dir = tmp_trestle_dir / 'catalogs/'
    rep_file = catalogs_dir / rep_name / 'catalog.json'

    # execute the command to replicate the model into replicated
    test_args = f'trestle replicate catalog -n {source_name} -o {rep_name} {regen}'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    rc = Trestle().run()
    assert rc == 0

    # now load the replicate and compare

    rep_model_type, rep_model_alias, rep_model_instance = ModelUtils.load_distributed(rep_file, tmp_trestle_dir)

    expected_model_type, _ = ModelUtils.get_stripped_model_type(rep_file.resolve(), tmp_trestle_dir)

    expected_model_instance = Catalog.oscal_read(testdata_dir / 'split_merge/load_distributed/catalog.json')

    assert rep_model_type == expected_model_type
    assert rep_model_alias == 'catalog'
    # last modified should be different
    assert expected_model_instance.metadata.last_modified != rep_model_instance.metadata.last_modified
    # set last modified equal so that equality can pass
    expected_model_instance.metadata.last_modified = rep_model_instance.metadata.last_modified
    assert (expected_model_instance == rep_model_instance) == (regen == '')


@pytest.mark.parametrize('regen', ['', '-r'])
def test_replicate_cmd_failures(testdata_dir, tmp_trestle_dir, regen, monkeypatch: MonkeyPatch) -> None:
    """Test replicate command failure paths."""

    def load_distributed_permission_error(*args, **kwargs):
        raise PermissionError

    def load_distributed_error(*args, **kwargs):
        raise err.TrestleError('load_distributed_error')

    def load_distributed_return(*args, **kwargs):
        model_instance = Catalog.oscal_read(local_data_file)
        return any, 'catalog', model_instance

    def execute_trestle_error(*args, **kwargs):
        raise err.TrestleError('execute_trestle_error')

    # prepare trestle project dir with the file
    test_utils.ensure_trestle_config_dir(tmp_trestle_dir)
    source_name = 'minimal_catalog'
    test_data_file = testdata_dir / 'json' / (source_name + '.json')
    catalogs_dir = Path('catalogs/')
    local_data_file = catalogs_dir / source_name / 'catalog.json'
    (catalogs_dir / source_name).mkdir()
    shutil.copy(test_data_file, local_data_file)
    rep_name = 'repcatalog'
    rep_file = catalogs_dir / rep_name / 'catalog.json'
    shutil.rmtree(catalogs_dir / rep_name, ignore_errors=True)
    (catalogs_dir / rep_name).mkdir()

    # now create pre-existing replica to force error
    rep_file.touch()

    test_args = f'trestle replicate catalog -n {source_name} -o {rep_name} -tr {tmp_trestle_dir}'.split()
    monkeypatch.setattr(sys, 'argv', test_args)
    rc = Trestle().run()
    assert rc == 1

    shutil.rmtree(catalogs_dir / rep_name, ignore_errors=True)

    args = argparse.Namespace(
        trestle_root=tmp_trestle_dir, name=source_name, output=rep_name, verbose=0, regenerate=False
    )

    # Force PermissionError:
    monkeypatch.setattr(
        'trestle.core.commands.replicate.ModelUtils.load_distributed', load_distributed_permission_error
    )
    with pytest.raises(PermissionError):
        ReplicateCmd.replicate_object('catalog', args)

    # Force TrestleError only:
    monkeypatch.setattr('trestle.core.commands.replicate.ModelUtils.load_distributed', load_distributed_error)
    with pytest.raises(err.TrestleError):
        ReplicateCmd.replicate_object('catalog', args)

    monkeypatch.setattr('trestle.core.commands.replicate.ModelUtils.load_distributed', load_distributed_return)
    monkeypatch.setattr('trestle.core.commands.replicate.Plan.rollback', mock_return)
    monkeypatch.setattr('trestle.core.commands.replicate.Plan.execute', execute_trestle_error)
    with pytest.raises(err.TrestleError):
        ReplicateCmd.replicate_object('catalog', args)

    # Force TrestleError in simulate:
    monkeypatch.setattr('trestle.core.commands.replicate.ModelUtils.load_distributed', load_distributed_return)
    monkeypatch.setattr('trestle.core.commands.replicate.Plan.execute', execute_trestle_error)
    with pytest.raises(err.TrestleError):
        ReplicateCmd.replicate_object('catalog', args)


def test_replicate_load_file_failure(tmp_trestle_dir: Path, monkeypatch: MonkeyPatch) -> None:
    """Test model load failures."""
    test_utils.ensure_trestle_config_dir(tmp_trestle_dir)

    # Create a file with bad json
    sample_data = '"star": {'
    source_dir = Path('catalogs/bad_catalog')
    source_dir.mkdir(exist_ok=True)
    bad_file_path = source_dir / 'catalog.json'
    bad_file = bad_file_path.open('w+', encoding=const.FILE_ENCODING)
    bad_file.write(sample_data)
    bad_file.close()

    args = 'trestle replicate catalog -n bad_catalog -o rep_bad_catalog'.split()
    monkeypatch.setattr(sys, 'argv', args)
    rc = Trestle().run()
    assert rc == 1


def test_replicate_file_system(tmp_trestle_dir: Path, monkeypatch: MonkeyPatch) -> None:
    """Test model load failures."""
    test_utils.ensure_trestle_config_dir(tmp_trestle_dir)

    args = argparse.Namespace(trestle_root=tmp_trestle_dir, name='foo', output='bar', verbose=0)
    monkeypatch.setattr('trestle.common.file_utils.extract_trestle_project_root', mock_return)
    with pytest.raises(err.TrestleError):
        ReplicateCmd.replicate_object('catalog', args)



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OSCAL transformation tasks."""

import configparser
import json
import logging
import pathlib
import traceback
from typing import Optional

from ruamel.yaml import YAML

from trestle.oscal.profile import Profile
from trestle.tasks.base_task import TaskBase
from trestle.tasks.base_task import TaskOutcome
from trestle.transforms.implementations.osco import OscalProfileToOscoProfileTransformer

logger = logging.getLogger(__name__)


class ProfileToOsco(TaskBase):
    """
    Task to convert Profile to OSC yaml.

    Attributes:
        name: Name of the task.
    """

    name = 'oscal-profile-to-osco-profile'

    def __init__(self, config_object: Optional[configparser.SectionProxy]) -> None:
        """
        Initialize trestle task oscal-profile-to-osco-profile.

        Args:
            config_object: Config section associated with the task.
        """
        super().__init__(config_object)

    def print_info(self) -> None:
        """Print the help string."""
        logger.info(f'Help information for {self.name} task.')
        logger.info('')
        logger.info(
            'Purpose: Transform Open Security Controls Assessment Language (OSCAL) Profile '
            + 'into Open Shift Compliance Operator (OSCO) .yaml file.'
        )
        logger.info('')
        logger.info('Configuration flags sit under [task.oscal-profile-to-osco-profile]:')
        logger.info('  input-file = (required) path of the input file comprising OSCAL profile.')
        logger.info('  output-dir = (required) path of the output directory comprising synthesized .yaml file.')
        logger.info(
            '  output-name = (optional) name of created file in output directory, default is osco-profile.yaml.'
        )
        logger.info('  output-overwrite = (optional) true [default] or false; replace existing output when true.')
        logger.info(
            '  quiet = (optional) true or false [default]; display file creations and rules analysis when false.'
        )
        logger.info('')
        logger.info('Operation: The specified input profile is transformed into OSCO .yaml.')
        logger.info('')
        logger.info('Notes:')
        note11 = '[1] The input-file OSCAL profile should specify a metadata property with'
        note12 = 'name "osco_version" and value of the form "0.1.46".'
        note13 = 'The value corresponds with the OpenShift Compliance Operator (OSCO) version'
        note14 = 'and affects the format of the emitted yaml.'
        note15 = 'If not specified, the default is "0.1.46".'
        logger.info(f'{note11} {note12} {note13} {note14} {note15}')
        note21 = '[2] For OSCO version "0.1.39" and prior no "description" is emitted for "spec".'
        logger.info(f'{note21}')

    def simulate(self) -> TaskOutcome:
        """Provide a simulated outcome."""
        return TaskOutcome('simulated-success')

    def execute(self) -> TaskOutcome:
        """Provide an actual outcome."""
        try:
            return self._execute()
        except Exception:
            logger.warning(traceback.format_exc())
            return TaskOutcome('failure')

    def _execute(self) -> TaskOutcome:
        """Perform transformation."""
        # check config
        if not self._config:
            logger.warning('config missing')
            return TaskOutcome('failure')
        # input-file
        input_file = self._config.get('input-file')
        if input_file is None:
            logger.warning('config missing "input-file"')
            return TaskOutcome('failure')
        logger.info(f'input-file: {input_file}')
        input_path = pathlib.Path(input_file)
        # output-dir
        output_dir = self._config.get('output-dir')
        if output_dir is None:
            logger.warning('config missing "output-dir"')
            return TaskOutcome('failure')
        output_path = pathlib.Path(output_dir)
        # insure output dir exists
        output_path.mkdir(exist_ok=True, parents=True)
        # output file path
        output_name = self._config.get('output-name', 'osco-profile.yaml')
        output_filepath = pathlib.Path(output_dir, output_name)
        logger.info(f'output-file: {output_filepath}')
        # overwrite
        overwrite = self._config.getboolean('output-overwrite', True)
        if not overwrite and pathlib.Path(output_filepath).exists():
            logger.warning(f'output-file: {output_filepath} already exists')
            return TaskOutcome('failure')
        # read input
        profile = Profile.oscal_read(input_path)
        # transform
        transformer = OscalProfileToOscoProfileTransformer()
        ydata = json.loads(transformer.transform(profile))
        # write output
        yaml = YAML(typ='safe')
        yaml.default_flow_style = False
        with open(output_filepath, 'w') as outfile:
            yaml.dump(ydata, outfile)
        # success
        return TaskOutcome('success')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A control markdown node."""
from __future__ import annotations

import logging
import re
import string
from enum import Enum
from typing import Dict, List, Optional, Tuple

from trestle.common import const
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_filtered_list, as_list, none_if_empty
from trestle.common.str_utils import spaces_and_caps_to_snake
from trestle.core.control_interface import ControlInterface
from trestle.core.markdown.base_markdown_node import BaseMarkdownNode, BaseSectionContent
from trestle.core.markdown.markdown_const import CODEBLOCK_DEF, HTML_COMMENT_END_REGEX, HTML_COMMENT_START
from trestle.oscal import common

logger = logging.getLogger(__name__)


class ControlSectionType(Enum):
    """A type of the section in the markdown.

    Supported types:
    1 - Control Statement
    2 - Control Objective
    3 - Control Guidance
    4 - ## Part foo
    5 - Control foo
    6 - ### foo
    7 - # Editable content

    Matching is case-insensitive.
    """

    STATEMENT = 1
    OBJECTIVE = 2
    GUIDANCE = 3
    EDITABLE_BY_ID_PART = 4
    GENERIC_CONTROL_PART = 5
    GENERIC_SUBPART = 6
    EDITABLE_CONTENT = 7
    UNDEFINED = 8


class TreeContext:
    """A shared singleton tree context."""

    def __init__(self):
        """Initialize tree context."""
        self.control_id = ''
        self.control_group = ''
        self.control_title = ''
        self.part_label_to_id_map = None
        self.section_to_part_name_map = None

    def reset(self):
        """Reset global control tree context."""
        self.control_id = ''
        self.control_group = ''
        self.control_title = ''
        self.part_label_to_id_map = None
        self.section_to_part_name_map = None


tree_context = TreeContext()


class ControlSectionContent(BaseSectionContent):
    """A content of the node."""

    def __init__(self):
        """Initialize section content.

        Attributes:
            part: A part that is found in markdown. A part is defined in markdown by two or more # symbols
            by_id_name: Required for parts defined as ## Part a in a markdown
        """
        super(ControlSectionContent, self).__init__()
        self.part = None
        self.by_id_name = ''

    def union(self, node: ControlMarkdownNode) -> None:
        """Unites contents together."""
        super().union(node)


class ControlMarkdownNode(BaseMarkdownNode):
    """Markdown will be read to the tree."""

    def __init__(self, key: str, content: ControlSectionContent, starting_line: int):
        """Initialize markdown node."""
        super(ControlMarkdownNode, self).__init__(key, content, starting_line)
        self.content: ControlSectionContent = content

    def _build_tree(self,
                    lines: List[str],
                    root_key: str,
                    starting_line: int,
                    level: int,
                    parent_part_id: str = '') -> Tuple[ControlMarkdownNode, int]:
        """
        Build a tree from the markdown recursively.

        The tree is contructed with valid headers as node's keys
        and node's content contains everything that is under that header.
        The subsections are placed into node's children with the same structure.

        A header is valid iff the line starts with # and it is not:
          1. Inside of the html blocks
          2. Inside single lined in the <> tags
          3. Inside the html comment
          4. Inside any table, code block or blockquotes

        Specific Control Rules:
          1. If we are in control statement or objective, no subsections are allowed.
        """
        content = ControlSectionContent()
        node_children = []
        i = starting_line

        section_heading_type = self._get_section_heading_type(root_key)

        part_id, part_name, part_title, by_id_part = self._get_part_info_from_section_name(root_key, parent_part_id,
                                                                                           section_heading_type)

        current_key_lvl = self._get_header_level_if_valid(root_key)

        if section_heading_type == ControlSectionType.UNDEFINED and tree_context.control_id:
            logger.warning(
                f'Undefined section {root_key} is found in the markdown for control {tree_context.control_id}. '
                f'This section will be ignored. Please make sure the spelling is correct.'
            )

        if current_key_lvl and current_key_lvl == 1 and section_heading_type != ControlSectionType.EDITABLE_CONTENT:
            # Parse control title
            if tree_context.control_id:
                logger.debug(
                    f'Duplicate control_id is found for the markdown {root_key}, '
                    f'make sure you have reset tree context before reading another markdown.'
                    f'Use markdown processor to avoid this error.'
                )
                raise TrestleError(
                    f'Multiple top-level headers are found but only one header is allowed. See line {root_key} '
                    f'for control {tree_context.control_id}.'
                )
            tree_context.control_id, tree_context.control_group, tree_context.control_title = ControlMarkdownNode._parse_control_title_line(root_key)  # noqa: E501
            content.control_id = tree_context.control_id
            content.control_group = tree_context.control_group
            content.control_title = tree_context.control_title

        def _strip_prose_or_none(part: Optional[common.Part]):
            if part and part.prose:
                part.prose = part.prose.strip() if part.prose.strip() else None

        def _add_child_prose_if_need(part: common.Part, text: str, should_add: bool):
            if should_add:
                part.prose += text

        while i < len(lines):
            line = lines[i].strip(' ')
            header_lvl = self._get_header_level_if_valid(line)

            if header_lvl is not None:
                if header_lvl >= level + 1:
                    if section_heading_type in [ControlSectionType.STATEMENT, ControlSectionType.OBJECTIVE]:
                        raise TrestleError(
                            f'Control Statement or Objective sections cannot contain subsections but found: {line}.'
                            f'Please delete this subsection and refer to docs on the required format.'
                        )
                    # build subtree
                    subtree, i = self._build_tree(lines, line, i + 1, level + 1, part_id)
                    node_children.append(subtree)
                    content.union(subtree)
                    if content.part and subtree.content.part and section_heading_type not in [
                            ControlSectionType.STATEMENT, ControlSectionType.OBJECTIVE, ControlSectionType.GUIDANCE
                    ]:
                        # Control statement, objective and guidance have special treatment
                        # in those sections any subsections go to the prose rather than subparts.
                        content.part.parts = as_list(content.part.parts)
                        content.part.parts.append(subtree.content.part)
                    # Control parts can have general markdown in the prose, if we are in the control part
                    # add its contents under the prose of a parent
                    _add_child_prose_if_need(
                        content.part,
                        subtree.content.raw_text,
                        section_heading_type
                        in [ControlSectionType.STATEMENT, ControlSectionType.OBJECTIVE, ControlSectionType.GUIDANCE]
                    )
                    continue
                else:
                    i -= 1  # need to revert back one line to properly handle next heading
                    break  # level of the header is above or equal to the current level, subtree is over

            if section_heading_type not in [ControlSectionType.UNDEFINED, ControlSectionType.EDITABLE_CONTENT]:
                # Read part
                content.part = self._create_part_if_needed(content.part, part_name, part_id)
                content.part.title = part_title
                read_parts = section_heading_type in [ControlSectionType.STATEMENT, ControlSectionType.OBJECTIVE]
                if by_id_part and section_heading_type == ControlSectionType.EDITABLE_BY_ID_PART:
                    content.by_id_name = by_id_part
                i = self._process_part_line(i, line, lines, content.part, read_parts=read_parts)
                continue

            # Nothing to do, simply increment
            i += 1

        first_line_to_grab = starting_line - 1 if starting_line else 0
        content.raw_text = '\n'.join(lines[first_line_to_grab:i])

        _strip_prose_or_none(content.part)

        md_node = ControlMarkdownNode(key=root_key, content=content, starting_line=first_line_to_grab)
        md_node.subnodes = node_children
        return md_node, i

    def get_control_statement(self) -> Optional[ControlMarkdownNode]:
        """Get control statement node."""
        return self.get_node_for_key('## Control Statement')

    def get_control_objective(self) -> Optional[ControlMarkdownNode]:
        """Get control objective node."""
        return self.get_node_for_key(const.CONTROL_OBJECTIVE_HEADER)

    def get_control_guidance(self) -> Optional[ControlMarkdownNode]:
        """Get control guidance node."""
        return self.get_node_for_key('## Control Guidance')

    def get_by_id_parts(self) -> Dict[str, List[common.Part]]:
        """Get by id editable parts."""
        part_id_to_parts_map = {}
        for node_key in as_filtered_list(self.content.subnodes_keys, lambda k: re.match(const.PART_REGEX, k)):
            # A by id part section
            part_node = self.get_node_for_key(node_key)
            if not part_node.content.part:
                raise TrestleError(f'Error no part was found in section {part_node.key}.')
            if part_node.content.by_id_name not in part_id_to_parts_map:
                part_id_to_parts_map[part_node.content.by_id_name] = []
            for subpart in part_node.subnodes:
                # We only care about subsections of ## Part section
                if subpart.content.part:
                    part_id_to_parts_map[part_node.content.by_id_name].append(subpart.content.part)

        return part_id_to_parts_map

    def get_editable_parts_and_subparts(self) -> Optional[List[common.Part]]:
        """Get editable parts and subparts if exist."""
        by_id_parts = self.get_by_id_parts()
        exclude_parts_id = list(by_id_parts.keys())
        editable_node = self.get_node_for_key(f'# {const.EDITABLE_CONTENT}')
        parts = self._get_subparts(editable_node, exclude_parts_id)
        return parts

    def _get_subparts(self,
                      control_node: ControlMarkdownNode,
                      exclude_ids: List[str] = None) -> Optional[List[common.Part]]:
        """Get subparts of the control part node if exists."""
        if not control_node:
            raise TrestleError(
                'No control node was provided to extract subparts. '
                'Please make sure your markdown contains # Editable Content section.'
            )
        parts = []
        for editable_part in as_filtered_list(control_node.subnodes, lambda p: p.content.part is not None):
            if editable_part.content.part.id not in exclude_ids:
                parts.append(editable_part.content.part)
        return parts

    def _get_section_heading_type(self, root_key: str) -> ControlSectionType:
        """Determine the section type based on the heading."""
        if root_key.lower() == '## control statement':
            return ControlSectionType.STATEMENT
        elif root_key.lower() == '## control objective':
            return ControlSectionType.OBJECTIVE
        elif root_key.lower() == '## control guidance':
            return ControlSectionType.GUIDANCE
        elif self._does_contain(root_key, const.PART_REGEX):
            return ControlSectionType.EDITABLE_BY_ID_PART
        elif f'# {const.EDITABLE_CONTENT}'.lower() in root_key.lower():
            return ControlSectionType.EDITABLE_CONTENT
        elif self._does_contain(root_key, const.CONTROL_REGEX):
            return ControlSectionType.GENERIC_CONTROL_PART
        elif '### ' in root_key.lower():
            return ControlSectionType.GENERIC_SUBPART

        return ControlSectionType.UNDEFINED

    def _get_part_info_from_section_name(self, root_key: str, parent_id: str,
                                         section_heading_type: ControlSectionType) -> Tuple[str, str, str, str]:
        """Get part information such as id, name and title based on the section heading."""
        part_id = ''
        part_name = ''
        part_title = None
        by_part_id = None  # special case used for ## Part

        if section_heading_type in [ControlSectionType.GUIDANCE, ControlSectionType.GENERIC_CONTROL_PART]:
            prefix = const.CONTROL_HEADER + ' '
            control_md_heading_label = root_key[len(prefix):].strip()
            control_md_heading_label_ncname = ControlInterface.strip_to_make_ncname(control_md_heading_label)
            control_md_heading_label_snakename = spaces_and_caps_to_snake(control_md_heading_label)

        if section_heading_type == ControlSectionType.STATEMENT:
            part_id = ControlInterface.create_statement_id(tree_context.control_id)
            part_name = const.STATEMENT

        if section_heading_type == ControlSectionType.OBJECTIVE:
            part_id = f'{tree_context.control_id}_obj'
            part_name = 'objective'

        if section_heading_type == ControlSectionType.GUIDANCE:
            # Read control guidance to a part object
            part_id = ControlInterface.strip_to_make_ncname(tree_context.control_id + '_gdn')
            part_name = control_md_heading_label_ncname

        if section_heading_type == ControlSectionType.EDITABLE_BY_ID_PART:
            # Read editable part
            by_part_label = re.match(const.PART_REGEX, root_key.lower()).groups(0)[0]
            control_label_map = tree_context.part_label_to_id_map.get(tree_context.control_id, None)
            if control_label_map is None:
                raise TrestleError(f'No label map found for control {tree_context.control_id}')
            by_part_id = control_label_map.get(by_part_label, None)
            if by_part_id is None:
                raise TrestleError(f'No part id found for label {by_part_label} in control {tree_context.control_id}')

            part_name = spaces_and_caps_to_snake(root_key.replace('#', '').replace('.', '').strip())
            part_id = f'{by_part_id}'
            part_title = root_key.replace('#', '').strip()

        if section_heading_type == ControlSectionType.GENERIC_CONTROL_PART:
            # Read other control parts to a part objects
            part_name = control_md_heading_label_snakename
            if tree_context.section_to_part_name_map:
                if control_md_heading_label in tree_context.section_to_part_name_map:
                    part_name = tree_context.section_to_part_name_map[control_md_heading_label]

                part_id = f'{tree_context.control_id}_{part_name}'
                part_title = control_md_heading_label
            else:
                part_id = spaces_and_caps_to_snake(tree_context.control_id + '_' + control_md_heading_label)

        if section_heading_type == ControlSectionType.GENERIC_SUBPART:
            # Read other control parts to a part objects
            match = re.match(const.AFTER_HASHES_REGEX, root_key)
            if not match:
                raise TrestleError(f'Unexpected editable header {root_key} found in part {part_id}')
            part_name_raw = match.groups(0)[0]
            part_name = spaces_and_caps_to_snake(part_name_raw)
            parent_suffix = parent_id + '.' if parent_id else ''
            part_id = parent_suffix + part_name

        return part_id, part_name, part_title, by_part_id

    def get_other_control_parts(self) -> List[Optional[ControlMarkdownNode]]:
        """Get all other control parts that are not statement, guidance or objective."""
        all_other_nodes = []
        all_control_sections = self.get_all_headers_for_key(const.CONTROL_HEADER, False)
        control_statement = self.get_control_statement()
        control_objective = self.get_control_objective()
        control_guidance = self.get_control_guidance()
        control_statement_heading = control_statement.key if control_statement else ''
        control_objective_heading = control_objective.key if control_objective else ''
        control_guidance_heading = control_guidance.key if control_guidance else ''
        for heading_key in all_control_sections:
            if heading_key not in {control_statement_heading, control_objective_heading, control_guidance_heading}:
                section_node = self.get_node_for_key(heading_key)
                all_other_nodes.append(section_node)

        return all_other_nodes

    def _create_part_if_needed(self, content_part: common.Part, part_name: str, part_id: str, prose: str = ''):
        """Create a new part if does not exist or return existing part."""
        if not content_part:
            if not tree_context.control_id:
                raise TrestleError(
                    f'Unexpected error, control id, group and title should be before ## Control {part_name}.'
                    'However, none was found.'
                )

            content_part = common.Part(name=part_name, id=part_id, prose=prose)

        return content_part

    def _process_part_line(
        self, line_idx: int, line: str, lines: List[str], part: common.Part, read_parts: bool = True
    ) -> int:
        """
        Process line for the part.

        If the read_parts is given and the line starts with '-' then
        the markdown list will be read to the subparts of the given part.
        """
        if self._does_start_with(line, HTML_COMMENT_START):
            comment_lines, line_idx = self._read_html_block(lines, line, line_idx + 1, HTML_COMMENT_END_REGEX)
            if line_idx >= len(lines):
                return line_idx
            line = lines[line_idx]

        if self._does_start_with(line, CODEBLOCK_DEF):
            code_lines, line_idx = self._read_code_lines(lines, line, line_idx + 1)
            if line_idx >= len(lines):
                return line_idx
            line = lines[line_idx]
            part.prose += '\n'.join(code_lines)

        if not line or not line.lstrip():
            # Empty line
            part.prose += '\n'
            line_idx += 1
            return line_idx

        if line.lstrip()[0] != '-' or not read_parts:
            # Line of text in prose
            part.prose += line + '\n'
            line_idx += 1
        else:
            # A part of inside of statement part
            if read_parts:
                end_idx, parts = self._read_parts(0, line_idx, lines, part.id, [])
                part.parts = none_if_empty(parts)
                line_idx = end_idx
            else:
                logger.warning(
                    f'{part.name} does not support subparts, ignoring {line} in control {tree_context.control_id}.'
                )
                line_idx += 1

        return line_idx

    def _read_parts(self, indent: int, ii: int, lines: List[str], parent_id: str,
                    parts: List[common.Part]) -> Tuple[int, List[common.Part]]:
        """If indentation level goes up or down, create new list or close current one."""
        while True:
            ii, new_indent, line = ControlMarkdownNode._get_next_indent(ii, lines)
            if new_indent < 0:
                # we are done reading control statement
                return ii, parts
            if new_indent == indent:
                # create new item part and add to current list of parts
                id_text, prose = ControlMarkdownNode._read_part_id_prose(line)
                # id_text is the part id and needs to be as a label property value
                # if none is there then create one from previous part, or use default
                if not id_text:
                    prev_label = ControlInterface.get_label(parts[-1]) if parts else ''
                    id_text = ControlMarkdownNode._create_next_label(prev_label, indent)
                id_ = ControlInterface.strip_to_make_ncname(parent_id.rstrip('.') + '.' + id_text.strip('.'))
                name = const.OBJECTIVE_PART if id_.find('_obj') > 0 else const.ITEM
                prop = common.Property(name='label', value=id_text)
                part = common.Part(name=name, id=id_, prose=prose, props=[prop])
                if id_ in [p.id for p in parts]:
                    logger.warning(
                        f'Duplicate part id {id_} is found in markdown '
                        f'{tree_context.control_id}. Please correct the part label in line {line}.'
                    )
                parts.append(part)
                ii += 1
            elif new_indent > indent:
                # add new list of parts to last part and continue
                if len(parts) == 0:
                    raise TrestleError(f'Improper indentation structure: {line}')
                ii, new_parts = self._read_parts(new_indent, ii, lines, parts[-1].id, [])
                if new_parts:
                    parts[-1].parts = new_parts
            else:
                # return list of sub-parts
                return ii, parts

    @staticmethod
    def _indent(line: str) -> int:
        """Measure indent of non-empty line."""
        if not line:
            raise TrestleError('Empty line queried for indent.')
        if line[0] not in [' ', '-', '\t']:
            return -1

        indent = 0
        for ii in range(len(line)):
            if line[ii] == '-':
                return indent
            # if line is indented it must start with -
            if line[ii] == '\t':
                # treat tab as two spaces
                indent += 2
            elif line[ii] != ' ':
                break
            else:
                indent += 1
        raise TrestleError(f'List elements must start with -: {line}')

    @staticmethod
    def _parse_control_title_line(line: str) -> Tuple[str, str, str]:
        """Process the title line and extract the control id, group title (in brackets) and control title."""
        if line.count('-') == 0:
            raise TrestleError(f'Markdown control title format error, missing - after control id: {line}')
        split_line = line.split()
        if len(split_line) < 3 or split_line[2] != '-':
            raise TrestleError(f'Cannot parse control markdown title for control_id group and title: {line}')
        # first token after the #
        control_id = split_line[1]
        group_title_start = line.find('\[')
        group_title_end = line.find('\]')
        if group_title_start < 0 or group_title_end < 0 or group_title_start > group_title_end:
            raise TrestleError(f'unable to read group title for control {control_id}')
        group_title = line[group_title_start + 2:group_title_end].strip()
        control_title = line[group_title_end + 2:].strip()
        return control_id, group_title, control_title

    @staticmethod
    def _read_part_id_prose(line: str) -> Tuple[str, str]:
        """Extract the part id letter or number and prose from line."""
        start = line.find('\\[')
        end = line.find('\\]')
        prose = line.strip() if start < 0 else line[end + 2:].strip()
        id_ = '' if start < 0 or end < 0 else line[start + 2:end]
        return id_, prose

    @staticmethod
    def _create_next_label(prev_label: str, indent: int) -> str:
        """
        Create new label at indent level based on previous label if available.

        If previous label is available, make this the next one in the sequence.
        Otherwise start with a or 1 on alternate levels of indentation.
        If alphabetic label reaches z, next one is aa.
        Numeric ranges from 1 to 9, then 10 etc.
        """
        if not prev_label:
            # assume indent goes in steps of 2
            return ['a', '1'][(indent // 2) % 2]
        label_prefix = ''
        label_suffix = prev_label
        is_char = prev_label[-1] in string.ascii_letters
        # if it isn't ending in letter or digit just append 'a' to end
        if not is_char and prev_label[-1] not in string.digits:
            return prev_label + 'a'
        # break in middle of string if mixed types
        if len(prev_label) > 1:
            ii = len(prev_label) - 1
            while ii >= 0:
                if prev_label[ii] not in string.ascii_letters + string.digits:
                    break
                if (prev_label[ii] in string.ascii_letters) != is_char:
                    break
                ii -= 1
            if ii >= 0:
                label_prefix = prev_label[:(ii + 1)]
                label_suffix = prev_label[(ii + 1):]

        return label_prefix + ControlMarkdownNode._bump_label(label_suffix)

    @staticmethod
    def _bump_label(label: str) -> str:
        """
        Find next label given a string of 1 or more pure letters or digits.

        The input must be either a string of digits or a string of ascii letters - or empty string.
        """
        if not label:
            return 'a'
        if label[0] in string.digits:
            return str(int(label) + 1)
        if len(label) == 1 and label[0].lower() < 'z':
            return chr(ord(label[0]) + 1)
        # if this happens to be a string of letters, force it lowercase and bump
        label = label.lower()
        factor = 1
        value = 0
        # delta is needed because a counts as 0 when first value on right, but 1 for all others
        delta = 0
        for letter in label[::-1]:
            value += (ord(letter) - ord('a') + delta) * factor
            factor *= 26
            delta = 1

        value += 1

        new_label = ''
        delta = 0
        while value > 0:
            new_label += chr(ord('a') + value % 26 - delta)
            value = value // 26
            delta = 1
        return new_label[::-1]

    @staticmethod
    def _get_next_indent(ii: int, lines: List[str], skip_empty_lines: bool = True) -> Tuple[int, int, str]:
        """Seek to next content line.  ii remains at line read."""
        while 0 <= ii < len(lines):
            line = lines[ii]
            if line and not line.isspace():
                if line[0] == '#':
                    return ii, -1, line
                indent = ControlMarkdownNode._indent(line)
                if indent >= 0:
                    # extract text after -
                    start = indent + 1
                    while start < len(line) and line[start] == ' ':
                        start += 1
                    if start >= len(line):
                        raise TrestleError(f'Invalid line {line}')
                    return ii, indent, line[start:]
                return ii, indent, line
            elif not skip_empty_lines:
                return ii, -1, line
            ii += 1
        return ii, -1, ''
